/*****************************************************************************/
//    Backtrack.C
/*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "Arg.h"
#include "VarArg.h"
#include "ExprArg.h"
#include "GroupingArg.h"
#include "PredAnnotations.h"
#include "Globals.h"
#include "RuleInfo.h"
#include "Literal.h"
#include "RelationMisc.h"
#include "BuiltinSymbols.h"
#include "Relation.h"

/*****************************************************************************/

extern void corannotation_error(corPredAnnotations *anno, char *msg);


static void check_this_arg(corCArg *, struct corVarLink *); // only affect VarArg
static int delete_this_arg(corCArg *a, struct corVarLink * varlist);


static void read_vars_from_arglist(struct corVarLink* varlist, corArgList& args)
{
    FOR_EACH_ARG(a, args) {
        if (a->IsEmpty()) {
            fputs("Something wrong here!\n",corexEnv->error_file);
            continue;
        }
	if (a->Kind() == OTHER_ARG) 
	  check_this_arg(a, varlist);
    } END_EACH_ARG
}

/*****************************************************************************/
static void check_this_arg(corCArg* a, struct corVarLink* varlist)
{
    struct corVarLink *temp;

    switch(a->kindof()) {
    case COR_VARIABLE:
        temp = new corVarLink(a->VarArgP(), varlist->next);
        varlist->next = temp;
        break;
    case COR_FUNCTOR:
        read_vars_from_arglist(varlist, (a->FuncArgP()->args));
        break;
    case COR_GROUPING:
	read_vars_from_arglist(varlist, a->GroupingP()->nested_args());
        break;
    case COR_NUM_CONST:
        if ( a->cNum_kindof() == COR_ARITH_EXPR) {
            check_this_arg(&(a->ExprArgP()->left_arg), varlist);
            check_this_arg(&(a->ExprArgP()->right_arg), varlist);
        }
        break;
    default:
        break;
    }
}


/*****************************************************************************/
// returns non-zero if any delete occurs, and 0 otherwise.

static int delete_vars_from_arglist(struct corVarLink* varlist, corArgList& args)
{
    int ret_val = 0;

    FOR_EACH_ARG(a, args) {
        ret_val = ret_val | delete_this_arg(a,varlist);
    } END_EACH_ARG;

    return ret_val;
}

/*****************************************************************************/
static int delete_this_arg(corCArg *a, struct corVarLink * varlist)
{
    int ret_val = 0;
    struct corVarLink *temp, *temp1;

    switch(a->kindof()) {
    case COR_VARIABLE: 
        for (temp = varlist; temp->next; temp = temp->next) {
            if (a->equals(temp->next->arg)) {
                temp1 = temp->next;
                temp->next = temp1->next;
                delete temp1;
                return 1;
            }
        }
        break;

    case COR_FUNCTOR:
        return delete_vars_from_arglist(varlist, (a->FuncArgP()->args));

    case COR_NUM_CONST:
        if ( a->cNum_kindof() == COR_ARITH_EXPR) {
            ret_val = delete_this_arg(& (a->ExprArgP()->left_arg), varlist);
            ret_val =
                ret_val | delete_this_arg(&(a->ExprArgP()->right_arg),
                         varlist);
            return ret_val;
        }

    case COR_GROUPING:
        break;

    default:			
        break;
    }
    return 0;
}

/*****************************************************************************/
/*
    Never backtrack to relations that can succeed only once. Examples of such
    relations are most builtins. Also, do not backtrack to a negated literal.
    In both these cases, go back to the previous literal.
    i is the current backtrack position chosen
*/

static int skipSingleSuccess(corRuleInfo *rInfo, int i, int *predOrders = NULL)
{
    int j;

    if (i < 0) return i;

    corRelation *rel;
    while (i >= 0) {
        if (predOrders)
            j = predOrders[i];
        else
            j = i;

        rel = corfind_external_relation(rInfo->rule->preds[j]->name(),
            rInfo->rule->preds[j]->arity());

        if (rel && (rel->single_success)) {
            // handles builtins
            i--;
        }
        else {
            // handles negated literals
            if (NEGATED(j))
                i--;
            else
                break;
        }
    }

    return i;
}

/*****************************************************************************/
/*
    If a cut has been crossed in the process of backtracking, then the
    backtrack point must be -1.
        i = current backtrack position chosen
        src = position for which i is the backtrack position
*/

static int cutCrossed(corRuleInfo *rInfo, int i, int src)
{
    if (i < 0) return i;

    int j;

    for (j = i; j < src; j++) 
        if (rInfo->rule->preds[j]->name() == corCutSymbol) break;

    if (j == src)
        return i;
    else
        return -1;
}


/*****************************************************************************/
/*
    success backtracking : is invoked upon succesful evaluation of
    a rule. Execution backtracks to the right-most literal that generates
    new bindings for the head literal. This function returns the success
    backtracking point for the seminaive rule.
    If there is a choice annotation on the head which specifies that
    (X) -> (Y), where (X) and (Y) are disjoint sets of head variables,
    then upon success, execution should backtrack to the right-most literal
    that generates new bindings for those head variables other than ones in(Y).
*/

int corgetSuccessBacktrackPoint(corSemiNaive *sn, corClause *clause, int ng_facts)
{
    corPredAnnotations *cur;
    int i = -1;

    if (ng_facts) {
        i = sn->rInfo->num_literals-1;
    }
    else {
        struct corVarLink *head_vars = new corVarLink(NULL, NULL);

        // Read the variables from the head into a VarLink list
        read_vars_from_arglist(head_vars, sn->rInfo->
            arg_list[sn->rInfo->num_literals]);


        if (head_vars->next) {
            for (i = -1; i < sn->rInfo->num_literals-1; ) {
                for( cur=corparserStruct.agg_sel_annotations ;cur!=NULL;
                    cur = cur->next())
                {
                    if (cur->pred == clause->head->pred) {
                        if (cur->op_arg.kindof() != COR_FUNCTOR) {
                            corannotation_error(cur, "Bad aggregation type");
                            break;
                        }
                        corFuncArg *fop_arg = cur->op_arg.FuncArgP();
                        if (fop_arg->functor() == corAnyOpSymbol) {
                            corVarLink *gbvars = new corVarLink(NULL,NULL);
                            read_vars_from_arglist(gbvars, *(cur->arglist2));
                            // Get the groupby vars
        
                            for(int j=0;j<=i; j++) {
                                delete_vars_from_arglist(gbvars,
                                    sn->rInfo->arg_list[sn->predOrders[j]]);
                                if (!(gbvars->next)) break;
                            }
                            if (!(gbvars->next))
                                delete_vars_from_arglist(head_vars,
                                    fop_arg->args);
                            // All vars in groupby list have been bound -- can
                             // now delete the variables in the "any"
                            // selection's head from the list of head_vars.
                        }
                        if (!(head_vars->next))
                            break; /* From inner for loop */
                    }
                }
                if(!(head_vars->next))
                    break;
                i++;
                delete_vars_from_arglist(head_vars,
                    sn->rInfo->arg_list[sn->predOrders[i]]);
    
                if(!(head_vars->next))
                    break;
            }
        }

        while (head_vars) {
            struct corVarLink *temp = head_vars->next;
            delete head_vars;
            head_vars = temp;
        }
    }
    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.

    if (i < 0) return i;
    
    return skipSingleSuccess(sn->rInfo, i, sn->predOrders);
}

/*****************************************************************************/
int corgetSuccessBacktrackPoint(corRuleInfo *rInfo, int ng_facts)
{
    int i = -1;

    if (ng_facts) {
        i = rInfo->num_literals-1;
    }
    else {
        struct corVarLink *head_vars = new corVarLink(NULL, NULL);
        int last_binding_point = -1;

        // Read the variables from the head into a VarLink list
        read_vars_from_arglist(head_vars,
            rInfo->arg_list[rInfo->num_literals]);

        if (head_vars->next) {
            for (i = 0; i < rInfo->num_literals; i++) {
                delete_vars_from_arglist(head_vars,
                    rInfo->arg_list[i]);
                last_binding_point = i;
                
                if(!(head_vars->next))
                    break;
            }
        }
        
        while (head_vars) {
            struct corVarLink *temp = head_vars->next;
            delete head_vars;
            head_vars = temp;
        }

        i = last_binding_point;
    }
        
    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.
    
    int j = skipSingleSuccess(rInfo, i);

    // check if a cut has been crossed
    return cutCrossed(rInfo, j, rInfo->num_literals);
}

/*****************************************************************************/
/*
 * get-first-failure backtracking : is invoked when a rhs literal
 * fails the first time it is reached. In this case, execution should
 * backtrack to the rightmost point that generates fresh bindings for
 * it. Example : p(X,Y,C) :- q(X, Z), r(Z,A,B), s(Z,C), t(A,B,C).
 * if s(Z,C) fails the first time it is reached, execution should
 * backtrack to q(X,Z), not to r(Z,A,B). However, if the failure
 * occurs after s(Z,C) has returned succesfully a few times, it is
 * WRONG to backtrack to q(X,Z). r(Z,A,B) could generate some new bindings for
 * A and B that might result in an answer tuple.
 */

int corgetfirstFailureBacktrackPoint(corSemiNaive *sn, int rhs_index, int ng_facts)
{
    int i = -1;

    if (ng_facts) {
        i = rhs_index-1;
    }
    else {
        struct corVarLink *rhs_literal_vars = new corVarLink(NULL, NULL);
        int last_binding_point = -1;
        
        // Read the variables from the rhs literal into a VarLink list
        read_vars_from_arglist(rhs_literal_vars,
            sn->rInfo->arg_list[sn->predOrders[rhs_index]]);

        if (rhs_literal_vars->next) {
            for (i = 0; i < rhs_index; i++) {
                if (delete_vars_from_arglist(rhs_literal_vars,
                    sn->rInfo->arg_list[sn->predOrders[i]]))
                {
                    last_binding_point = i;
                }
    
                if (!(rhs_literal_vars->next))
                    break;
            }
        }
        
        while (rhs_literal_vars) {
            struct corVarLink *temp = rhs_literal_vars->next;
            delete rhs_literal_vars;
            rhs_literal_vars = temp;
        }

        i = last_binding_point;
    }

    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.

    if (i < 0) return i;

    return skipSingleSuccess(sn->rInfo, i, sn->predOrders);
}

/*****************************************************************************/
/*
 * get-next-failure backtracking : is invoked when a rhs literal
 * fails after it has been successful at least once. 
 * In this case, execution should backtrack to the rightmost point that
 * generates fresh bindings for the literal or any literal after it in the
 * rule execution order.
 * Example : p(X,Y,C) :- q(X, Z), r(A,B), s(Z,C), t(A,B,C).
 * if s(Z,C) fails when it is reached on a get-next, execution should
 * backtrack to r(A,B), not to q(X,Z)).
 */

int corgetnextFailureBacktrackPoint(corSemiNaive *sn, int rhs_index, int ng_facts)
{
    int i = -1;

    if (ng_facts) {
        i = rhs_index-1;
    }
    else {
        struct corVarLink *rhs_literal_vars = new corVarLink(NULL, NULL);
        int last_binding_point = -1;

        // Read the variables from the rhs literal into a VarLink list
        for (int tmp = rhs_index; tmp < sn->rInfo->num_literals; tmp++)

        read_vars_from_arglist(rhs_literal_vars,
             sn->rInfo->arg_list[sn->predOrders[tmp]]);

        if (rhs_literal_vars->next) {
            for (i = 0; i < rhs_index; i++) {
                if (delete_vars_from_arglist(rhs_literal_vars,
                    sn->rInfo->arg_list[sn->predOrders[i]]))
                {
                    last_binding_point = i;
                }
    
                if(!(rhs_literal_vars->next))
                    break;
            }
        }
        
        while (rhs_literal_vars) {
            struct corVarLink *temp = rhs_literal_vars->next;
            delete rhs_literal_vars;
            rhs_literal_vars = temp;
        }
        
        i = last_binding_point;
    }
    
    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.

    if (i < 0) return i;
    
    return skipSingleSuccess(sn->rInfo, i, sn->predOrders);
}

/*****************************************************************************/
int corgetfirstFailureBacktrackPoint(corRuleInfo *rInfo, int rhs_index, int ng_facts)
{
    int i = -1;

    if (ng_facts) {
        i = rhs_index-1;
    }
    else {
        struct corVarLink *rhs_literal_vars = new corVarLink(NULL, NULL);
        int last_binding_point = -1;

        // Read the variables from the rhs literal into a VarLink list
        read_vars_from_arglist(rhs_literal_vars,
            rInfo->arg_list[rhs_index]);
        
        if (rhs_literal_vars->next) {
            for (i = 0; i < rhs_index; i++) {
                if (delete_vars_from_arglist(rhs_literal_vars,
                    rInfo->arg_list[i]))
                {
                    last_binding_point = i;
                }
    
                if(!(rhs_literal_vars->next))
                    break;
            }
        }
        
        while (rhs_literal_vars) {
            struct corVarLink *temp = rhs_literal_vars->next;
            delete rhs_literal_vars;
            rhs_literal_vars = temp;
        }
        
        i = last_binding_point;
    }
    
    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.
    
    int j = skipSingleSuccess(rInfo, i);

    // check if a cut has been crossed
    return cutCrossed(rInfo, j, rhs_index);
}

/*****************************************************************************/
int corgetnextFailureBacktrackPoint(corRuleInfo *rInfo, int rhs_index, int ng_facts)
{
    int i = -1;

    if (ng_facts) {
        i = rhs_index-1;
    }
    else {
        struct corVarLink *rhs_literal_vars = new corVarLink(NULL, NULL);
        int last_binding_point = -1;
        
        // Read the variables from the rhs literal into a VarLink list
        for (int tmp = rhs_index; tmp < rInfo->num_literals; tmp++) {
            read_vars_from_arglist(rhs_literal_vars,
                rInfo->arg_list[tmp]);
        }
        
        if (rhs_literal_vars->next) {
            for (i = 0; i < rhs_index; i++) {
                if (delete_vars_from_arglist(rhs_literal_vars,
                    rInfo->arg_list[i]))
                {
                    last_binding_point = i;
                }
    
                if(!(rhs_literal_vars->next))
                    break;
            }
        }
        
        while (rhs_literal_vars) {
            struct corVarLink *temp = rhs_literal_vars->next;
            delete rhs_literal_vars;
            rhs_literal_vars = temp;
        }
        
        i = last_binding_point;
    }

    //    At this stage, i is the index of the rhs literal at which forward
    //    computation should resume in the rule after a successful firing.
    
    int j = skipSingleSuccess(rInfo, i);

    // check if a cut has been crossed
    return cutCrossed(rInfo, j, rhs_index);
}

/*****************************************************************************/

