/*****************************************************************************/
//	CRdbRelation.C
/*****************************************************************************/

#include "CRdbManager.h"

#ifdef DO_RDB

// #include <stdio.h>
#include <assert.h>
// #include <string.h>
// #include <ctype.h>
#include "CRdbRelation.h"
#include "Interface.h"
#include "Database.h"
#include "Globals.h"
#include "CoralError.h"
#include "FuncArg.h"
#include "SymbolTable.h"
#include "CoralStream.h"
#include "BoundArgList.h"
#include "Log.h"
#include "String.h"


/*****************************************************************************/

// make this more global...
inline ostream& operator <<(ostream& os, const RdbItemList& items)
{
    char sep = '(';
    for( int i = 0 ; i < items.Size() ; i++ ) {
	os << sep << items[i];
	sep = ',';
    }
    return os << ')';
}


/*****************************************************************************/

// this should really be the constructor, & it could delete the
// previous version
int CreateCRdbRelation(Name newName, RdbJoin& join)
{
    Association* pAssoc = corSymbolLookup(corCurDB->RelationTable,newName);
    if(!HashNone(pAssoc->arg) ) // name exists!
      return COR_BAD_ARGLIST;

    corCRdbRelation* rel = new corCRdbRelation(newName,join);
    pAssoc->arg = newName;
    pAssoc->val = rel;
    // Insert the new relation in the relation table.
    corTabInserted(corCurDB->RelationTable, pAssoc);
    return 0;
}

/*****************************************************************************/

corCRdbRelation::corCRdbRelation(Name newName, const RdbJoin& newJoin)
: corRelation(),
  joinInfo(newJoin)
{
    // register with the manager
    // pCRdbManager->StoreCRdbRelation(this);
    assert( joinInfo.RelCount() > 0 );
    name = newName;
    r_kind = (RelationKind) COR_R_RDBREL;
    queryVersions = new RdbQueryVersions(joinInfo);
    _arity = queryVersions->Arity();
    rdbInsert = NULL;
    deleteVersions = NULL;
}

/*****************************************************************************/

corCRdbRelation::~corCRdbRelation()
{
    if( deleteVersions ) {
	delete deleteVersions;
	deleteVersions = NULL;
    }
    if( rdbInsert ) {
	delete rdbInsert;
	rdbInsert = NULL;
    }
    if( queryVersions ) {
	delete queryVersions;
	queryVersions = NULL;
    }
    // remove from manager
    //pCRdbManager->RemoveCRdbRelation(this);
}

/*****************************************************************************/

RdbCursor* corCRdbRelation::Init(corTupleIterator& iterator)
{
    RdbCursor* csr = NULL;
    if ((iterator.ipos == Scanning_Init)) {
	csr = Bind( iterator );
	iterator.stack_mark.get_mark();
	iterator.ipos = Scanning_Known;
	iterator.scanDesc = csr;
    } else {
	iterator.stack_mark.pop_to();
	csr = (RdbCursor*)iterator.scanDesc;
    }
    return csr;
}

/*****************************************************************************/

corTuple* corCRdbRelation::Done(corTupleIterator& iterator)
{
    if( iterator.scanDesc ) {               
	delete (RdbCursor*)iterator.scanDesc;
	iterator.scanDesc = NULL;
    }
    iterator.ipos = Scanning_Done;
    iterator.set_no_match();
    return NULL;
}

/*****************************************************************************/

corTuple* corCRdbRelation::get_next_tuple(corTupleIterator& iterator)
{
    if(corexEnv->C_interrupt_raised)
      return Done(iterator);

    RdbCursor* csr = Init(iterator);
    if( !csr )
      return Done(iterator);

    RdbItemList result;

    // keep getting tuples from the rdbms until:
    //   a. there are no more (eof)
    //   b. there is an error
    //   c. the tuple unifies with the iterator - should this be done??
    while( !csr->GetNext(result) ) {

	LOG( "Rec returned: " << result << endl );

	// convert the RdbItemList to an ArgList
	corArgList* arglist = AsArgList(result);

	// if the tuple unifies, return it
	corTuple* tuple = new corTuple(arglist);
	if( corunify_literal(iterator,tuple,corexEnv->dbg_get_next) )
	  return tuple;

	// tuple did not unify, clean up and get another one.
	delete tuple;
	delete arglist;	
    }
    // eof or error
    return Done(iterator);
}

/*****************************************************************************/

RdbCursor* corCRdbRelation::Bind(corTupleIterator& iterator)
{
    RdbItemList bindings;
    BitVector adorn;
    if( AsRdbItemList(iterator.arg_list, iterator.bindenv, bindings, adorn) )
      return NULL;
    RdbQuery& query = queryVersions->Lookup(adorn);
    return new RdbCursor(query, bindings);
}

/*****************************************************************************/

// returns 0 on success, 1 on error
int corCRdbRelation::AsRdbItemList(corArgList& argList,
				corBindEnv* pBindEnv,
				RdbItemList& result,
				BitVector& adorn) const
{
    int arity = argList.arity();
    adorn.setlen(arity);	// adornment of bindings
    for( int i = 0 ; i < arity ; i++ ) {
	corTerm term(argList[i], pBindEnv);
	cordeep_dereference_term(term);
	switch( term.expr.kindof() ) {
	  case COR_SYMBOL_CONST:
	    {
		adorn.set(i);
		char* str = corSymbolString(term.expr);
		result.Append( RdbItem(str) );
		break;
	    }
	  case COR_NUM_CONST:
	    {
		adorn.set(i);
		corCArg numarg = term.expr;
		switch( numarg.cNum_kindof() ) {
		  case COR_INTEGER:
		    result.Append(RdbItem(long(numarg.NumIval())));
		    break;
		  case COR_DOUBLE:
		    result.Append(RdbItem(numarg.NumDval()));
		    break;
		  default:
		    assert(0);	// unknown type...
		}
	    }
	  case COR_VARIABLE:
	    // do nothing
	    break;
	  default:
	    CORAL_error(COR_BAD_ARG_TYPE, NULL, "CRdbRelation::AsRdbItemList");
	    return 1;
	}
    }
    return 0;
}

/*****************************************************************************/

corArgList* corCRdbRelation::AsArgList(RdbItemList& result)
{
    corArgList* arglist = new corArgList(result.Size());
    for( int i = 0 ; i < result.Size() ; i++ ) {
	switch( result[i].Type() ) {
	  case RdbItem::RDB_STRING:
	    {
		String str = String(result[i]); 
		str.RTrim();
		(*arglist)[i] = make_arg((char*)(str));
		break;
	    }
	  case RdbItem::RDB_INT:
	    (*arglist)[i] = corCArg(int(RdbInt(result[i])));
	    break;
	  case RdbItem::RDB_REAL:
	    (*arglist)[i] = corCArg(double(RdbReal(result[i])));
	    break;
	  case RdbItem::RDB_NULL:
	    {
		char* na = "(null)";
		// char* s = new char[ strlen(na) +1 ];
		// strcpy(s,na);
		(*arglist)[i] = make_arg(na);
		break;
	    }
	  default:
	    {
		assert(0);
	    }
	}
    }
    return arglist;
}

/*****************************************************************************/

void corCRdbRelation::release( corTupleIterator& iterator )
{
    Done(iterator);
}

/*****************************************************************************/

int corCRdbRelation::insert_tuple(corTuple* newTuple)
{
    return insert_new(newTuple->args(), newTuple->bindenv);
}

/*****************************************************************************/

// check subsumption, then inserts into backend dbms
int corCRdbRelation::insert_new(corArgList& argList, corBindEnv* pBindEnv, 
			     corBindEnv* , corTuple* , int )
{
    if(corexEnv->C_interrupt_raised) return 0;  

    TRACE_INSERTION("[Starting to insert " << corSymbolString(name)
		    << '(' << corBoundArgList(argList,pBindEnv) << ')' );

    if( joinInfo.RelCount() != 1 ) {
	ERROR_MSG(" - attempt to insert tuple into a join of rdb relations\n");
	TRACE_INSERTION("insertion failed]\n");
        return 0;
    }

    if( !rdbInsert ) {		// allocate if first try to insert
	rdbInsert = joinInfo.Rel(0).NewInsert();
    }

    if( argList.arity() != arity() ) {
	ERROR_MSG("attempt to insert tuple of arity " << argList.arity()
		  << "\n\t\t- in relation " << name->string()
		  << " of arity " <<  arity() << endl);
	TRACE_INSERTION("insertion failed]\n");
        return 0;
    }

    RdbItemList tuple;
    BitVector adorn(arity());

    if( AsRdbItemList(argList, pBindEnv, tuple, adorn) ) {
	// error reported in AsRdbItemList
	TRACE_INSERTION("insertion failed]\n");
	return 0;
    }

    if( adorn.count() != arity() ) { // all of the args must be bound
	ERROR_MSG("attempt to insert a non-ground tuple\n");
	TRACE_INSERTION("insertion failed]\n");
        return 0;
    }
	
    // subsumption (duplicate) checks
    if (check_subsum==1) {
	RdbQuery& qry = queryVersions->Lookup(adorn); // fully bound
	RdbCursor csr(qry, tuple);
	RdbItemList result;
	if( !csr.GetNext(result) ) { // tuple found, subsume
	    // this is not an error
	    TRACE_INSERTION("-- subsumed]\n");
	    return 0;
	}
    }

    int rc = rdbInsert->Insert(tuple);
    if( rc ) {
	ERROR_MSG("received rdb error " << rc << endl);
	TRACE_INSERTION("insertion failed]\n");
	return 0;
    }

    TRACE_INSERTION("-- done]\n");
    return 1;			// successfully added a new row
}

/*****************************************************************************/

// delete all tuples that match 'tuple'
int corCRdbRelation::tuple_delete(corTuple* tuple)
{
    if(corexEnv->C_interrupt_raised) return 0;  
    
    if( joinInfo.RelCount() != 1 ) {
	ERROR_MSG("attempt to delete tuple into a join of rdb relations\n");
        return 0;
    }

    if( !deleteVersions ) {	// allocate if this is the first time
	deleteVersions = new RdbDeleteVersions(joinInfo.Rel(0));
    }

    RdbItemList rdbTuple;
    BitVector adorn(tuple->arity());
    if( AsRdbItemList(tuple->args(), tuple->bindenv, rdbTuple, adorn) ) {
	// error already reported
	return 0;
    }

    RdbDelete& del = deleteVersions->Lookup(adorn);
    RdbError::ErrorCode rc = del.Delete(rdbTuple);
    if( rc == RdbError::SQL_EOF ) {
	// no matching rows found!
	return 0;
    }
    if( rc ) {
	ERROR_MSG("received rdb error " << rc << endl);
	return 0;
    }

    return 1;			// successfully added a new row
}

/*****************************************************************************/

void corCRdbRelation::print_name( FILE* outf )
{
    String sys = queryVersions->SystemName();
    String db = queryVersions->DatabaseName();
    String schema = queryVersions->Schema();
    fprintf(outf, "%15s/%d\t: %s/%s:%s\n", corSymbolString(name), 
	    arity(), (char*)sys, (char*)db, (char*)schema );
}

/*****************************************************************************/

// takes a funcarg like: coralname(rel1(X,Y),rel2(Y,Z)...)
int CreateCRdbJoinRelation(corFuncArg* pfaNewRel, corBindEnv* pBindEnv)
{
    int relCount = pfaNewRel->arity();
    if(relCount < 1)		// must select at least one rel
      return COR_BAD_ARGLIST;

    RdbRelList relList;
    RdbJoinCondList joinList;
    corCRdbVarPtrList varList;

    // foreach relation to be joined
    for( int rel = 0 ; rel < relCount ; rel++ ) {
	corTerm tRel(pfaNewRel->args[rel], pBindEnv);
	cordeep_dereference_term(tRel);
	if(tRel.expr.kindof() != COR_FUNCTOR )
	  return COR_BAD_ARGLIST;
	
	corFuncArg* pfaRel = tRel.expr.FuncArgP();
	int arity = pfaRel->arity();
	if(arity < 1)		// must select at least one attr.
	  return COR_BAD_ARGLIST;

	corRelation* pRel = find_relation(pfaRel->functor(),arity);
	if( pRel == NULL )	// couldn't find relation
	  return COR_BAD_ARGLIST;
	
	if( pRel->r_kind != COR_R_RDBREL ) // not an RDB relation
	  return COR_BAD_ARGLIST;

	// make ptr to crdb subtype
	corCRdbRelation* pCRdbRel = (corCRdbRelation*)pRel; 

	// need to check that all are from the same system...
	// or should Rdb__ be responsible, probably.

	// should generalize this to work for joins too
	relList.Append(&pCRdbRel->RdbRel());

	// foreach attribute of this relation
	for( int attr = 0 ; attr < arity ; attr++ ) {
	    corTerm tAttr(pfaRel->args[attr], pBindEnv);
	    cordeep_dereference_term(tAttr);
	    if( tAttr.expr.kindof() == COR_VARIABLE ) {
		// why are don't cares getting in here??
		Name varName = (tAttr.expr.VarArgP())->var_name();
		if( strcmp(varName->string(), "_") != 0 ) { // skip don't cares
		    varList.Append(new corCRdbVar(rel,attr,varName));
		}
	    }
	}
    }
    CreateCRdbJoinList(joinList, varList);

    return CreateCRdbRelation(pfaNewRel->functor(), RdbJoin(relList,joinList));
}

/*****************************************************************************/

void CreateCRdbJoinList(RdbJoinCondList& joinList, corCRdbVarPtrList& varList)
{
    varList.SortByName();
    int count = varList.Size();
    for( int i = 1 ; i < count ; i++ ) {
	int j = i -1;
	while( j >= 0 && varList[i]->name == varList[j]->name ) {
	    joinList.Append( RdbJoinCondition(varList[i]->rel,	varList[i]->col,
					      varList[j]->rel, varList[j]->col));
	    j--;
	}
    }
}

/*****************************************************************************/

// implements insertion sort to put the list in order by name
void corCRdbVarPtrList::SortByName()
{
    corCRdbVar* t;
    int count = Size();
    for( int i = 1 ; i < count ; i++ ) {
	t = (*this)[i];
	int j = i;
	while( j > 0 && (*this)[j-1]->name > t->name ) {
	    Set(j,(*this)[j-1]); // want just a ptr swap, not struct
	    j--;
	}
	Set(j,t);
    }
}

/*****************************************************************************/
/*****************************************************************************/

#endif // DO_RDB

