/*****************************************************************************/
//	CRdbSolvers.C
/*****************************************************************************/

#include "CRdbManager.h"
#include "Util.h"

#ifdef DO_RDB

#include <string.h>
#include <strstream.h>
#include "CRdbSolvers.h"
#include "CRdbRelation.h"
#include "TupleIterator.h"
#include "SymbolTable.h"
#include "Database.h"
#include "Term.h"
#include "FuncArg.h"
#include "CoralError.h"
#include "Rdb.h"
#include "Log.h"


/*****************************************************************************/

// Standard error handling for solvers
#undef  ERR
#define ERR(A,B) { \
	CORAL_error((A),(B),fnName); \
	iterator.set_no_match(); \
	return iterator.bindenv; \
}

/*****************************************************************************/

corBindEnv * RdbMapSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* fnName = "RdbMapSolver";
    char* usage = "rdb_map(coral_dbname, coral_relname(relation("
      "col1,...,colN)),...)";
    int arg_count = iterator.arg_list.count();
    if( arg_count < 2 )
      ERR(COR_BAD_ARITY, usage);

    corTerm dbNameTerm (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(dbNameTerm);
    if( dbNameTerm.expr.kindof() != COR_SYMBOL_CONST )
      ERR(COR_BAD_ARGLIST, usage);
    char* dbName = corSymbolString(dbNameTerm.expr);

    // find the database
    RdbDatabase* pRdbDatabase = pCRdbManager->LookupDatabase(dbName);
    if( pRdbDatabase == NULL )
      ERR(COR_BAD_ARGLIST, "coral rdb database not open!");

    // foreach relation to be mapped
    for( int i = 1 ; i < arg_count ; i++ ) {
	// extract the relation name 
	corTerm coralNameTerm(iterator.arg_list[i], iterator.bindenv);
	cordeep_dereference_term(coralNameTerm);
	if (coralNameTerm.expr.kindof() != COR_FUNCTOR )
	  ERR(COR_BAD_ARGLIST, usage);
	corFuncArg* coralFunc = coralNameTerm.expr.FuncArgP();
	char* coralName = corSymbolString(coralFunc->functor());
	if (coralFunc->arity() != 1 ||
	    coralFunc->args[0].kindof() != COR_FUNCTOR )
	  ERR(COR_BAD_ARGLIST, usage);
	corFuncArg* relFunc = coralFunc->args[0].FuncArgP();
	char* relName = corSymbolString(relFunc->functor());

	// extract the attributes
	RdbColNameList cols;
	int arity = relFunc->arity();
	for( int j = 0 ; j < arity ; j++ ) {
	    corTerm colTerm (relFunc->args[j], iterator.bindenv);
	    cordeep_dereference_term(colTerm);
	    switch( colTerm.expr.kindof() ) {
	      case COR_SYMBOL_CONST:
		{
		    char* colName = corSymbolString(colTerm.expr);
		    cols.Append(String(colName));
		    break;
		}
	      default:
		ERR(COR_BAD_ARGLIST, usage);
	    }
	}

	// create the relation
	Name new_name = corEnterSymbol(coralName);
	Association *ptr = corSymbolLookup(corCurDB->RelationTable,new_name);
	if(!HashNone(ptr->arg)) // table name in use, error
	  ERR(COR_DUP_SCHEMA,"relation NOT replaced.");
	RdbRelation* rdbRel = new RdbRelation(*pRdbDatabase, String(relName),
					      cols);
	if( rdbRel->GetError() ) {
	    delete rdbRel;
	    ptr->arg = NULL;
	    ptr->val = NULL;
	    // should put in new COR_RDB errors...
	    char msg[1024];
	    ostrstream oss(msg,sizeof(msg)); //replace by stream error messages
	    oss << pRdbDatabase->SystemName() << "/"
	      << pRdbDatabase->DatabaseName() << ": " 
		<< pRdbDatabase->GetErrorMessage();
	    ERR(COR_RDB_ERR, msg);
	}
	RdbRelList relList;
	relList.Append(rdbRel);
	RdbJoinCondList joinCondList;
	RdbJoin join(relList,joinCondList);
	int rc = CreateCRdbRelation(new_name,join);
	if( rc )
	  ERR(rc, "could not create relation");
    } // for each relation to be mapped
    return iterator.bindenv ;
}

/***************************************************************************/

corBindEnv * RdbCommitSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* fnName = "RdbCommitSolver";
    char* usage = "rdb_commit([database])";
    int arg_count = iterator.arg_list.count();
    if( arg_count != 0 && arg_count != 1 )
      ERR(COR_BAD_ARITY, usage);

    if( arg_count == 0 ) {
	pCRdbManager->Commit();
    } else {
	corTerm dbNameTerm (iterator.arg_list[0], iterator.bindenv);
	cordeep_dereference_term(dbNameTerm);
	if( dbNameTerm.expr.kindof() != COR_SYMBOL_CONST )
	  ERR(COR_BAD_ARGLIST, usage);
	char* dbName = corSymbolString(dbNameTerm.expr);

	// find the rdbms system
	RdbDatabase* db = pCRdbManager->LookupDatabase(dbName);
	if( db == NULL )
	    ERR(COR_BAD_ARGLIST, "database not found!");

	db->Commit();
    }
    return iterator.bindenv;
}

/***************************************************************************/

corBindEnv * RdbRollbackSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* fnName = "RdbRollbackSolver";
    char* usage = "rdb_rollback([database])";
    int arg_count = iterator.arg_list.count();
    if( arg_count != 0 && arg_count != 1 )
      ERR(COR_BAD_ARITY, usage);

    // make this general...
    if( arg_count == 0 ) {
	pCRdbManager->Rollback();
    } else {
	corTerm dbNameTerm (iterator.arg_list[0], iterator.bindenv);
	cordeep_dereference_term(dbNameTerm);
	if( dbNameTerm.expr.kindof() != COR_SYMBOL_CONST )
	  ERR(COR_BAD_ARGLIST, usage);
	char* dbName = corSymbolString(dbNameTerm.expr);

	// find the rdbms system
	RdbDatabase* db = pCRdbManager->LookupDatabase(dbName);
	if( db == NULL )
	    ERR(COR_BAD_ARGLIST, "database not found!");

	db->Rollback();
    }
    return iterator.bindenv;
}

/***************************************************************************/

corBindEnv * RdbExecuteSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* fnName = "RdbExecuteSolver";
    char* usage = "rdb_execute(database,command)";
    int arg_count = iterator.arg_list.count();
    if( arg_count != 2 )
      ERR(COR_BAD_ARITY, usage);

    corTerm dbNameTerm (iterator.arg_list[0], iterator.bindenv);
    corTerm commandTerm (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term(dbNameTerm);
    cordeep_dereference_term(commandTerm);
    if( dbNameTerm.expr.kindof() != COR_SYMBOL_CONST 
	|| commandTerm.expr.kindof() != COR_SYMBOL_CONST )
      ERR(COR_BAD_ARGLIST, usage);

    char* dbName = corSymbolString(dbNameTerm.expr);
    char* cp = corSymbolString(commandTerm.expr);
    String command = cp;

    RdbDatabase* db = pCRdbManager->LookupDatabase(dbName);
    if( db == NULL )
      ERR(COR_BAD_ARGLIST, "database not found!");

    int rc = db->ImmediateSql(command);
    if( rc ) {
	ostrstream oss;
	oss << db->SystemName() << '/' << db->DatabaseName() << ": " 
	  << db->GetErrorMessage();
	char* msg = oss.str();
	ERR(COR_RDB_ERR, msg);
	delete [] msg;
    }
    return iterator.bindenv;
}

/*****************************************************************************/

corBindEnv * RdbJoinSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* usage = "rdb_join(newrel(rdbrel1(X,...),rdbrel2(X,...),...),...)";
    char* fnName = "RdbJoinSolver";
    int arg_count = iterator.arg_list.count();

    // for each new (join result) relation being defined
    for( int newRel = 0 ; newRel < arg_count ; newRel++ ) {
	corTerm tNewRel(iterator.arg_list[newRel], iterator.bindenv);
	cordeep_dereference_term(tNewRel);
	if(tNewRel.expr.kindof() != COR_FUNCTOR ) {
	    CORAL_error(COR_BAD_ARGLIST,usage,fnName);
	    continue;
	}
	
	corFuncArg* pfaNewRel = tNewRel.expr.FuncArgP();
	int rc = CreateCRdbJoinRelation(pfaNewRel,iterator.bindenv);
	if( rc ) {
	    CORAL_error(rc,usage,fnName);
	    continue;
	}
    }
    return iterator.bindenv;
}

/*****************************************************************************/

corBindEnv * RdbLoggingSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char* usage = "rdb_logging(on|off)";
    char* fnName = "RdbLoggingSolver";

#ifndef DO_LOGGING

    ERR(COR_BAD_ARGLIST, "Not compiled with logging!");

#else // DO_LOGGING

    int arg_count = iterator.arg_list.count();
    if( arg_count != 1)
      ERR(COR_BAD_ARGLIST, usage);

    corTerm tOpt(iterator.arg_list[0], iterator.bindenv);
    DEEP_DEREFERENCE_TERM(tOpt);
    if( tOpt.expr->kindof() != COR_SYMBOL_CONST )
      ERR(COR_BAD_ARGLIST, usage);

    char* coropt = corSymbolString(tOpt.expr);
    if( strcmp(coropt,"on") == 0 )
      klog.Enable();
    else if( strcmp(coropt,"off") == 0 )
      klog.Disable();
    else
      ERR(COR_BAD_ARGLIST, usage);

#endif // DO_LOGGING

    return iterator.bindenv;
}

/*****************************************************************************/

corBindEnv * RdbOpenDBSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    enum {			// the position of each argument
	WS_NAME, SYSTEM, DATABASE, USERID, PASSWORD, SERVER, 
	MAX_ARGC, REQ_ARGC = DATABASE+1, 
    };
    char* fnName = "RdbOpenDBSolver";
    char* usage = "rdb_open_db(ws_name, system, database, "
      "[, userid[, password [, server]]])";
    int argc = iterator.arg_list.count();
    if( argc < REQ_ARGC || argc > MAX_ARGC )
      ERR(COR_BAD_ARITY, usage);

    // extract parameters
    String argv[MAX_ARGC];	// each parameter
    for( int i = 0 ; i < argc ; i++ ) {
	corTerm term(iterator.arg_list[i], iterator.bindenv);
	cordeep_dereference_term(term);
	if( term.expr.kindof() != COR_SYMBOL_CONST )
	  ERR(COR_BAD_ARGLIST, usage);
	char* cp = corSymbolString(term.expr);
	argv[i] = cp;
    }

    // Check if the coral db name exists
    if( pCRdbManager->LookupDatabase(argv[WS_NAME]) != NULL )
      ERR(COR_BAD_ARGLIST, "coral database name is already in use!");

    // find the rdbms system
    RdbSystem* pRdbSystem = pCRdbManager->LookupSystem((char*)argv[SYSTEM]);
    if( pRdbSystem == NULL )
      ERR(COR_BAD_ARGLIST, "invalid system name!");

    // find the database
    RdbDatabase* pRdbDatabase = pRdbSystem->LookupDatabase(argv[DATABASE],
							   argv[SERVER]);
    if( pRdbDatabase == NULL )
      pRdbDatabase = pRdbSystem->OpenDatabase(argv[DATABASE], argv[USERID],
					      argv[PASSWORD], argv[SERVER]);
    if( pRdbDatabase == NULL )
      ERR(COR_BAD_ARGLIST, "rdb database could not be opened!");

    // eventually, I'm going to make this into an actual workspace...
    // create...DB(EnterSymbol(argv[WS_NAME]);

    pCRdbManager->StoreDatabase(argv[WS_NAME], pRdbDatabase);
    
    return iterator.bindenv;
}


/*****************************************************************************/
/*****************************************************************************/

#endif // DO_RDB


