// *************************************************************************
//    HashIndex.C
// *************************************************************************

#include "HashIndex.h"

#include "Globals.h"
#include "Execute.h"

// *************************************************************************

int corC_NewHashIndex = 0, corC_NewHashIndexSize = 0;
int corC_DeleteHashIndex = 0, corC_DeleteHashIndexSize = 0;

// *************************************************************************

extern int corsubsumes_arg_lists(corArgList& left, corArgList& right,
    corBindEnv *left_bindings, corBindEnv *right_bindings);
extern int corsubsumes_tuple(corTuple *tuple, corTuple *newtuple, int display);
int corunify_literal(corTupleIterator& iterator, corTuple *tuple, int display);

// *************************************************************************

#define BACKBONESIZE 16

// *************************************************************************
corHashIndex::corHashIndex(corStorageRelation* rel, corBitVector* aset, int buckets,
    double incr_ratio1, double max_occupancy1)
    : corGenericIndex( rel, aset )
{
    // for the 0-arity reln, or for the all-free index
    if ((rel->arity() == 0) || (!aset->any())) {
        buckets = 1;            // Ought to be 0, but too much code probably
                                // depends on a non-empty hash table.
        incr_ratio1 = 1.0;
        max_occupancy1 = 1000000.0;
    }

    if (buckets < 0) buckets = corexEnv->C_hash_buckets_default;
    if (incr_ratio1 < 0.0) incr_ratio1 = corexEnv->C_incr_ratio_default;
    if (max_occupancy1 < 0.0) max_occupancy1 = corexEnv->C_max_occupancy_default;

    hash_size = buckets;
    hash_count = 0;
    incr_ratio = incr_ratio1;
    max_occupancy = max_occupancy1;
    limit = (int) (hash_size * max_occupancy);

    hash_table = new HashNodePtr[hash_size];

    register int i;
    for (i = 0; i < hash_size; i++)
      hash_table[i] = NULL;
    var_chain = NULL;
    corC_NewHashIndex ++;
    corC_NewHashIndexSize += hash_size * sizeof(corHashNode *);

    non_empty_buckets = new int[hash_size];
    for (i = 0; i < hash_size; i++)
        non_empty_buckets[i] = -1;
    non_empty_bucket_head = -1 ;
    prev_bucket_ptrs = new int*[hash_size];
}

// *************************************************************************
corHashIndex::corHashIndex(corStorageRelation* rel, corArgList* pat, int n_var, 
    corArgList* v_nm, int buckets, double incr_ratio1, double max_occupancy1)
    : corGenericIndex(rel, pat, n_var, v_nm)
{
    if (rel->arity() == 0) {
        buckets = 1;    // Ought to be 0, but too much code probably depends on 
                          // a non-empty hash table.
    }

    if (buckets < 0) buckets = corexEnv->C_hash_buckets_default;
    if (incr_ratio1 < 0.0) incr_ratio1 = corexEnv->C_incr_ratio_default;
    if (max_occupancy1 < 0.0) max_occupancy1 = corexEnv->C_max_occupancy_default;

    hash_size = buckets;
    hash_count = 0;
    incr_ratio = incr_ratio1;
    max_occupancy = max_occupancy1;
    limit = (int) (hash_size * max_occupancy);

    hash_table = new HashNodePtr[hash_size];

    int i;
    for (i = 0; i < hash_size; i++)
        hash_table[i] = NULL;
    var_chain = NULL;
    corC_NewHashIndex++;
    corC_NewHashIndexSize += hash_size * sizeof(corHashNode *);

    non_empty_buckets = new int[buckets];
    for (i = 0; i < hash_size; i++)
        non_empty_buckets[i] = -1;
    non_empty_bucket_head = -1 ;
    prev_bucket_ptrs = new int*[buckets];

}

// *************************************************************************
void corHashIndex::reset_index()
{
    int i;

    non_empty_bucket_head = -1;
    var_chain = NULL;
    hash_count = 0;
    limit = (int) (hash_size *max_occupancy);
    for (i = 0; i < hash_size; i++)
        hash_table[i] = NULL;
    for (i = 0; i < hash_size; i++)
        non_empty_buckets[i] = -1;
}

// *************************************************************************
void corHashIndex::free_all_entries()
{
    register int j = 0;
    // delete nodes in the var_chain list
    HashNodePtr next;
    for (HashNodePtr h = var_chain; h; h = next) {
        next = h->next;
        delete h;
    }
    var_chain = NULL;

    hash_count = 0;

    // delete nodes in the hash buckets

    for (j=non_empty_bucket_head; j >= 0 ; j=non_empty_buckets[j]) {
        for (h = hash_table[j]; h; h = next) {
            next = h->next;
            delete h;
        }
        hash_table[j] = NULL;
    }

    non_empty_bucket_head = -1 ;
}

// *************************************************************************
void corHashIndex::move_all_entries(corGenericIndex *ind)
{
    if  (index_type() != ind->index_type()) {
        // WARNING - should also check for matching of indices: 
        // With the current implementation the indices will match,
        // but this is something that must be checked.  Insert a call :
        // if (!matches(ind)) error ....
        // It has been left out for efficiency reasons, and because 
        // matches is not fully implemented yet.
            
        fprintf(corexEnv->error_file,
            "CORAL:: Internal Error - move_all_entries type ");
        fprintf(corexEnv->error_file, "mismatch between indices.\n");
        return;
    }

    register corHashIndex *hindex = (corHashIndex *) ind;
    register HashNodePtr next, h;

    // Move nodes in the var_chain list
    for (h = var_chain; h; h = next) {
        next = h->next;
        hindex->insert_entry(h);
    }
    var_chain = NULL;

    hash_count = 0;

    // Move nodes in the hash buckets

    register int j;

    for (j=non_empty_bucket_head; j >= 0 ; j=non_empty_buckets[j]) {
        for (h = hash_table[j]; h; h = next) {
            next = h->next;
            hindex->insert_entry(h);
        }
        hash_table[j] = NULL;
    }
    non_empty_bucket_head = -1 ;
}

// *************************************************************************
void corHashIndex::delete_all_tuples()
{
#    ifdef DEBUG
        if (corexEnv->dbg_deletes) {
            fprintf(corexEnv->trace_file,
                "Deleting %d tuples from index of relation", hash_count);
            relation->name->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\n");
           }
#    endif

    // delete nodes in the var_chain list
    HashNodePtr next;
    for (HashNodePtr h = var_chain; h; h = next) {
        next = h->next;
        delete h->tuple();
        // h->value = NULL;
        delete h;
    }
    var_chain = NULL;

    hash_count = 0;

    // delete nodes in the hash buckets

    register int j ;
    for (j=non_empty_bucket_head; j >= 0 ; j=non_empty_buckets[j]) {
        for (h = hash_table[j]; h; h = next) {
            next = h->next;
            /*
            for (int ti = 0; ti < h->tuple()->arity(); ti++)
                // NUM_CONSTs can be safely deleted since they are never shared
                if (h->tuple()->arg(ti)->kindof() == COR_NUM_CONST)
                      delete h->tuple()->arg(ti);
            */
            delete h->tuple()->_args;
            delete h->tuple();
            delete h;
        }
        hash_table[j] = NULL;
    }
    non_empty_bucket_head = -1;
}

// *************************************************************************
corHashIndex::~corHashIndex()
{
    free_all_entries();
    corC_DeleteHashIndex++;
    corC_DeleteHashIndexSize += hash_size * sizeof(corHashNode *);

    delete [] non_empty_buckets;
    delete [] prev_bucket_ptrs;
  
    delete [] hash_table;
}

// *************************************************************************
/*
    The following is an optimization to let a static environment to be used in
    reentrant code.  We assume no parallelism however.
*/

static corArrayBindEnv static_env1(MAX_VARS,VAR_INCR);
static int static_env1_inuse=DONT_USE_STATIC_ENV;

// *************************************************************************
HashVal corHashIndex::hash(corArgList *args, corBindEnv *bindenv)
{
    if (args->count() == 0) {
        // Normally this doesn't really mean that arity==0.  Rather, it
        // means that the TupleIterator was created with an ArgList
        // parameter, which is a convention meaning essentially
        // (VAR1, VAR2, ..., VARn).  However, in pathological cases, it
        // might really be a nullary Relation. Do the same thing in either case.
        return VarHashValue;
    }

    if (argset != NULL) {
        // the argument form index will be used
        return args->hash(*argset, bindenv);
    }

    corStackMark stackmark0;
    HashVal result;

    // else the pattern form index will be used 

    corArrayBindEnv* pat_env;
    if (static_env1_inuse++ == 0 && num_vars_pat <= MAX_VARS) {
        pat_env = &static_env1;
    }
    else {
        pat_env = new corArrayBindEnv (num_vars_pat);
    }

    //    NOTE: a subsumption check alone is insufficient ! In the presence of
    //    non-ground facts, the tuple to insert may be more general than the
    //    pattern of the index, and unification is needed to ensure that
    //    all desired inserts occur --- PRAVEEN
    //    if (subsumes_arg_lists(*pattern, *args, pat_env, bindenv)) {

    if (corunify_arg_lists(*pattern, *args, pat_env, bindenv) == COR_U_SUCCEED) {
        corBitVector aset(var_names->count());
        for (int i = 0; i < var_names->count(); i++)
            aset.set(i);
        
        result = var_names->hash(aset, pat_env);    // hash on all arguments
        stackmark0.pop_to();
        if (--static_env1_inuse != 0 || num_vars_pat > MAX_VARS) {
            delete pat_env;
        }
        return result;
    } 

    // else the tuple and the pattern don't match
    stackmark0.pop_to();
    if (--static_env1_inuse != 0 || num_vars_pat > MAX_VARS) {
        delete pat_env;
    }
    return UnmatchedHashValue;
}

// *************************************************************************
void corHashIndex::insert_tuple( corTuple *tuple )
{
    if (tuple->is_deleted()) return;
    // tuple has been logically deleted

    HashVal h_val = hash(&tuple->args(), tuple->bindenv);

#    ifdef DO_TRACE
    if (corexEnv->dbg_indexing) {
        fprintf(corexEnv->trace_file, "HashIndex::insert_tuple: ");
        relation->name->print(NULL, corexEnv->trace_file);
        fprintf(corexEnv->trace_file,"(");
        tuple->printon(corexEnv->trace_file);
        fprintf(corexEnv->trace_file, ") ");

        if (argset != NULL) {
            // argument form index
            fprintf(corexEnv->trace_file, "[hash index on ");
            argset->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ]\n");
        }
        else {
            // pattern form index
            fprintf(corexEnv->trace_file, "[pattern matched ");
            pattern->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "] ");
               fprintf(corexEnv->trace_file, "[indexed on ");
            var_names->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ]\n");
        }
        fprintf(corexEnv->trace_file, "        hash value is %ld\n", h_val);
    }
#    endif

    corHashNode *new_node = new corHashNode(h_val, tuple); 
    tuple->ref_count++;
    if (h_val == VarHashValue) {
        // link it to the beginning of var_chain
        new_node->next = var_chain;
        var_chain = new_node;
#        ifdef DO_TRACE
            if (corexEnv->dbg_indexing) {
                fprintf(corexEnv->trace_file, "HashIndex::insert_tuple: ");
                fprintf(corexEnv->trace_file, "inserting in var_chain\n");
            }
#        endif
    }
    else if (h_val != UnmatchedHashValue) { 
        // link it to the beginning of the appropriate hash bucket

        hash_count ++;
        int tmp_hash_index = (int) (h_val % hash_size);
        new_node->next = hash_table[tmp_hash_index];
        hash_table[tmp_hash_index] = new_node;
        if (!new_node->next) {
            if (non_empty_bucket_head >= 0) {
                prev_bucket_ptrs[non_empty_bucket_head] =
                    &(non_empty_buckets[tmp_hash_index]);
            }
            non_empty_buckets[tmp_hash_index] = non_empty_bucket_head ;
            non_empty_bucket_head = tmp_hash_index ;
            prev_bucket_ptrs[tmp_hash_index] = &(non_empty_bucket_head);
        }

#        ifdef DO_TRACE
            if (corexEnv->dbg_indexing) {
                fprintf(corexEnv->trace_file, "HashIndex::insert_tuple: ");
                fprintf(corexEnv->trace_file, "inserting in %s bucket %d \n",
                    (new_node->next ? "non-empty" : "empty"), tmp_hash_index);
            }
#        endif
    }

    // else (h_val == UnmatchedHashValue) { } 
    // i.e, pattern does not match iterator

    if (hash_count > limit)
        rehash_bigger();
}

// *************************************************************************
void corHashIndex::insert_entry( HashNodePtr new_node)
{

    corTuple *tuple = new_node->tuple();
    HashVal h_val = new_node->hash_val;

    if (tuple->is_deleted()) {
        // tuple has been logically deleted
        return;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file, "HashIndex::insert_entry: ");
            relation->name->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file,"(");
            tuple->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ") ");

            if (argset != NULL) {
                // argument form index
                fprintf(corexEnv->trace_file, "[hash index on ");
                argset->printon(corexEnv->trace_file);
                fprintf(corexEnv->trace_file, " ]\n");
            }
            else {
                // pattern form index
                fprintf(corexEnv->trace_file, "[pattern matched ");
                pattern->print(NULL, corexEnv->trace_file);
                fprintf(corexEnv->trace_file, "] ");
                fprintf(corexEnv->trace_file, "[indexed on ");
                var_names->print(NULL, corexEnv->trace_file);
                fprintf(corexEnv->trace_file, " ]\n");
            }
            fprintf(corexEnv->trace_file, "        hash value is %ld\n", h_val);
        }
#    endif

    tuple->ref_count++;
    if (h_val == VarHashValue) {
        // link it to the beginning of var_chain
        new_node->next = var_chain;
        var_chain = new_node;

#        ifdef DO_TRACE
            if (corexEnv->dbg_indexing) {
                fprintf(corexEnv->trace_file, "HashIndex::insert_tuple: ");
                fprintf(corexEnv->trace_file, "inserting in var_chain\n");
            }
#        endif
    }
    else if (h_val != UnmatchedHashValue) { 
        // link it to the beginning of the appropriate hash bucket
        hash_count ++;
        int tmp_hash_index = (int)(h_val % hash_size);
        new_node->next = hash_table[tmp_hash_index];
        hash_table[tmp_hash_index] = new_node;

        if (!new_node->next) {
            if (non_empty_bucket_head >= 0) {
                prev_bucket_ptrs[non_empty_bucket_head] =
                    &(non_empty_buckets[tmp_hash_index]);
            }
            non_empty_buckets[tmp_hash_index] = non_empty_bucket_head ;
            non_empty_bucket_head = tmp_hash_index ;
            prev_bucket_ptrs[tmp_hash_index] = &(non_empty_bucket_head);
        }

#        ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file, "HashIndex::insert_tuple: ");
            fprintf(corexEnv->trace_file, "inserting in %s bucket %d \n",
                (new_node->next ? "non-empty" : "empty"), tmp_hash_index);
        }
#        endif
    }

    // else (h_val == UnmatchedHashValue) { } 
    // i.e, pattern does not match iterator

    if (hash_count > limit)
        rehash_bigger();
}

// *************************************************************************
void corHashIndex::rehash_bigger()
{
    int i;
    HashNodePtr cur, next, *prevptr;

    limit = (int) (limit *incr_ratio);
    int old_hash_size = hash_size;
    hash_size = (int)(hash_size * incr_ratio);

    delete non_empty_buckets;
    non_empty_buckets = new int[hash_size];
    for (i = 0; i < hash_size; i++)
        non_empty_buckets[i] = -1;
    non_empty_bucket_head = -1;
    delete prev_bucket_ptrs;
    prev_bucket_ptrs = new int*[hash_size];

    // Recalculate limit to avoid roundoff errors
    limit = (int) (hash_size * max_occupancy);

    HashNodePtr *old_table = hash_table;
    hash_table = new HashNodePtr[hash_size];
    for (i=0; i < hash_size; i++) {
        hash_table[i] = NULL;
    }

    for (i=0; i < old_hash_size; i++) {
        prevptr = &(old_table[i]);
        for(cur=*prevptr; cur != NULL;) {
            next = cur->next;
            int tmp_hash_index = (int)(cur->hash_val % hash_size);
            cur->next = hash_table[tmp_hash_index];
            hash_table[tmp_hash_index] = cur;
            if (!cur->next) {
                if (non_empty_bucket_head >= 0) {
                    prev_bucket_ptrs[non_empty_bucket_head] =
                        &(non_empty_buckets[tmp_hash_index]);
                }
                non_empty_buckets[tmp_hash_index] = non_empty_bucket_head ;
                non_empty_bucket_head = tmp_hash_index ;
                prev_bucket_ptrs[tmp_hash_index] = &(non_empty_bucket_head);
            }
            cur = next;
        }
    }

    delete old_table;

    corC_NewHashIndexSize += (hash_size - old_hash_size) * sizeof(corHashNode *);
}


// *************************************************************************
/*
    iterator.ipos is used to keep track of where were are in the scan.
    The possible values are:

        #define Scanning_Init  0      Initial state 
        #define Scanning_Known 1      scanning known bucket 
        #define Scanning_Vars  2      scanning from var_chain 
        #define Scanning_New   3      scanning from chain
        #define Scanning_Done  4      done
        #define Scanning_Hash  5
*/

// *************************************************************************
/*
    For i in [Scanning_Hash .. Scanning_Hash+hash_size-1] means
    that we are scanning all hash buckets, and are currently at
    bucket i-Scanning_Hash
*/

void corHashIndex::initialize_iterator(corTupleIterator& iterator)
{

  iterator.stack_mark.get_mark();    

  // hash() retruns VarHashValue if iterator.arg_list.count == 0
  iterator.hash = hash(&iterator.arg_list, iterator.bindenv); 
  // changed to hashing on the appropriate set of arguments 

#    ifdef DO_TRACE
    if (corexEnv->dbg_indexing) {
        fprintf(corexEnv->trace_file,
            "\nHashIndex::initialize_iterator: indexing ");
        iterator.relation->name->print(NULL, corexEnv->trace_file);
        fprintf(corexEnv->trace_file, " using ");
        iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
    
        if (argset != NULL) { 
            // argument form index
            fprintf(corexEnv->trace_file, "[hash index on ");
            argset->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ]\n");
        }
        else { 
            // pattern form index
            fprintf(corexEnv->trace_file, "[pattern matched ");
            pattern->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "] ");
            fprintf(corexEnv->trace_file, "[indexed on ");
            var_names->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ]\n");
        }
        fprintf(corexEnv->trace_file,
            "                 hash value is %d\n", iterator.hash);
    }
#    endif
    
    if (iterator.hash == UnmatchedHashValue) {
        iterator.ipos = Scanning_Done;
    }
    else if (iterator.hash != VarHashValue) {
        iterator.ipos = Scanning_Known;
        iterator.pindex = (int)(iterator.hash) % hash_size;
        iterator.ppos = &hash_table[iterator.pindex];
    } 
    else {
        // the entire relation has to be searched
        if (non_empty_bucket_head < 0) {
            iterator.ipos = Scanning_Vars;
            iterator.ppos = &var_chain;
            iterator.pindex = -1;
        }
        else {
            iterator.ipos = Scanning_Hash + non_empty_bucket_head;
            iterator.pindex = non_empty_bucket_head;
            iterator.ppos = &hash_table[iterator.pindex]; 
        }
    }
}

// *************************************************************************
/*
    HashIndex::get_next_tuple which returns a Tuple* and 
    HashIndex::get_next which returns a BindEnv* are extremely similar
    in their code.  In fact, get_next_tuple duplicates the entire
    code of get_next.  If trailing is used (i.e. DO_TRAIL is defined)
    get_next modifies the iterator.bindenv AND returns a pointer to 
    the BindEnv.  One way of making it consistent is to make it explicit
    that iterator.bindenv will contain the modified BindEnv, and 
    return a Tuple* consistently across the board.
*/

corTuple* corHashIndex::get_next_tuple(corTupleIterator& iterator)
{
    HashVal h_val;
    corBindEnv *result_env = iterator.bindenv;
    corStackMark  stackmark0;
    corTuple *result = NULL;
    iterator.tuple_env = NULL;
    
    if (iterator.ipos == Scanning_Init)
        initialize_iterator(iterator);
    else
        iterator.stack_mark.pop_to();
    
    for (;;) {
        // Continue until we find a match or hash-table exhausted
             
        HashNodePtr *next = (HashNodePtr*) iterator.ppos;
        if ((iterator.ipos == Scanning_Done) || (*next == NULL)) {
            switch (iterator.ipos) {
            case Scanning_Known:
                iterator.ipos = Scanning_Vars;
                iterator.ppos = &var_chain;
                continue;
        
            case Scanning_Vars:
            case Scanning_New: // this should not occur
            case Scanning_Done: 
                result = NULL;
                return result;
        
            default:
                // iterator.ipos >= Scanning_Hash
                h_val = iterator.ipos - Scanning_Hash;
                if (non_empty_buckets[(int)h_val] < 0) {
                    iterator.ipos = Scanning_Vars;
                    iterator.ppos = &var_chain;
                }
                else {
                    iterator.pindex = non_empty_buckets[(int)h_val];
                    iterator.ipos = Scanning_Hash + iterator.pindex;
                    iterator.ppos = &hash_table[iterator.pindex];
                }
                continue;
            }
        }

        HashNodePtr tuple_index = *next;
    
        if ((tuple_index->tuple())->is_deleted()) {
            // The tuple has been logically deleted catch the case where
            // iterator.ipos == ScanningKnown, and so iterator.pindex
            // becomes -4 (a wrong value)
            if (iterator.pindex < 0) {
                iterator.ppos = &tuple_index->next;
                continue;
            }

            if (relation->scan_count != 1) {
                iterator.ppos = &tuple_index->next;
                continue;
            }

            *(next) = (*next)->next;
            
            /*
                The if check is needed because we might have a rule of the form
                try :- p(X), delete(p(X)), print(X), p(Y), print(Y).
                The delete() builtin merely marks the tuples as deleted.
                The later p iterator must not reclaim the space, because this 
                will corrupt the first iterator.
            */

            // PRAVEEN:: I am doubtful about adding this here.  Will this break
            //    the system ?
            //    Looks like it is okay, except for ordered search. For some reason,
            //    in Ordered Search, a tuple is deleted from within insert_in_context
            //    and then crashes later here !!
            if (corCurrModuleInfo && !corCurrModuleInfo->UseOrdSearch &&
                !(--(tuple_index->tuple()->ref_count)))
            {
                /*
                for (int ti = 0; ti < tuple_index->tuple()->arity(); ti++) {
                    // NUM_CONSTs can be safely deleted since they are never shared
                    if (tuple_index->tuple()->arg(ti)->kindof() == COR_NUM_CONST)
                        delete tuple_index->tuple()->arg(ti);
                }
                */
                //if (tuple_index->tuple()->bindenv) 
                        //delete tuple_index->tuple()->bindenv;
                delete tuple_index->tuple()->_args;
                delete tuple_index->tuple();
                delete tuple_index;
            }

            if ((next == &hash_table[iterator.pindex]) && (*next == NULL)) {
                *(prev_bucket_ptrs[iterator.pindex]) =
                    non_empty_buckets[iterator.pindex];
                if (non_empty_buckets[iterator.pindex] >= 0) {
                    prev_bucket_ptrs[non_empty_buckets[iterator.pindex]] = 
                    prev_bucket_ptrs[iterator.pindex];
                }
            }
            hash_count--;
            
#            ifdef DO_PROFILE
                if (corT_Stack.count > 0) {
                    (corT_Stack.top()->moduledata->module_info.corGlobalProfileInfo.
                        deleted_tuples)++; 
                }
                else {
                    (corGlobalProfileInfo.deleted_tuples)++; 
                }
#            endif
            continue;
        }
    
        iterator.ppos = &tuple_index->next;
    
        // Since many hash values could be mapped to the same hash bucket,
        // this preliminary check is more efficient.
        if (iterator.ipos == Scanning_Known &&
            iterator.hash != tuple_index->hash_val) continue;
    
        if (!corCurrModuleInfo->NonGroundFacts) {
            int t_cnt = relation->arity();
            corTerm left_term, right_term;
            for (int l = 0; l < t_cnt; l++) {
                left_term.expr = iterator.arg_list[l];
                right_term.expr = tuple_index->tuple()->args()[l];
                left_term.bindenv = iterator.bindenv;
                right_term.bindenv = NULL;
                if (corunify_args(left_term, right_term) == COR_U_FAIL)
                    break;
            }
            if (l == t_cnt)
                return tuple_index->tuple();
            stackmark0.pop_to();
        }
        else {
            if (corunify_literal(iterator, tuple_index->tuple(),
                corexEnv->dbg_indexing))
            {
                return tuple_index->tuple();
            }
            stackmark0.pop_to();
        }
    }
}

// *************************************************************************
corBindEnv *corHashIndex::get_next(corTupleIterator& iterator)
{
    corTuple* tuple = get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    else
        return iterator.bindenv;
}

// *************************************************************************
int corHashIndex::is_subsumed(corTuple *newtuple)
{
    HashVal hash_val;
    HashNodePtr ppos;
    int ipos;
    corStackMark  stackmark0;
    /*
        ipos ==-1 indicates looking at varchain, else indicates number of
        buckets left to look at after current one
    */

    if (newtuple->args().count() == 0) {
        //    Normally this doesn't really mean that arity==0.
        //    Rather, it means that the TupleIterator was
        //    created with an ArgList parameter, which is a
        //    convention meaning essentially (VAR1, VAR2, ..., VARn).
        //    However, in pathological cases, it might really
        //    be a nullary Relation. Do the same thing in either case.
        hash_val = VarHashValue;
    }
    else {
        hash_val = (HashVal) hash(&(newtuple->args()), newtuple->bindenv); 
    }

    if (hash_val == UnmatchedHashValue) {
        return 0;
    }
    else if (hash_val != VarHashValue) {
        ipos = -2;            // Only one bucket to look at.
        ppos = hash_table[(int) hash_val % hash_size];
    }
    else {
        // the entire relation has to be searched
        ipos = non_empty_bucket_head;
        if (ipos >= 0)
            ppos = hash_table[ipos];
        else
            ppos = var_chain;
    }

    for (;ipos>=-2;) {
        // ipos == -1 indicates we are in varchain
        for(;ppos != NULL; ppos = ppos->next) {
            if ((ppos->tuple())->is_deleted())
                continue;    // the tuple has been logically deleted

            //    Since many hash values could be mapped to the same hash
            //    bucket, this preliminary check is more efficient.
            if ((hash_val != VarHashValue) &&
                (hash_val != ppos->hash_val) && 
                (ppos->hash_val != VarHashValue)) continue;
            
            if (corsubsumes_tuple(ppos->tuple(), newtuple, corexEnv->dbg_indexing))
                return 1;
        }
        if (ipos == -1) return 0;

        if (ipos == -2) {
            ipos = -1;
            ppos = var_chain;
        }
        else {
            ipos = non_empty_buckets[ipos];
            if (ipos >= 0) {
                ppos = hash_table[ipos];
            }
            else {
                if (ipos==-1) ppos = var_chain;
            }
        }
    }
    return 0;

}

// *************************************************************************
