/*****************************************************************************/
//    RuleInfo.C
/*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "RuleInfo.h"

#include "ArgList.h"
#include "StackMark.h"
#include "Config.h"
#include "Rule.h"
#include "Table.h"
#include "RMarkArray.h"
#include "TupleIterator.h"
#include "RelationList.h"
#include "SccInfo.h"
#include "VersionedBindEnv.h"
#include "Globals.h"
#include "Literal.h"
#include "Execute.h"
#include "CoralBreak.h"


/*****************************************************************************/

extern void corAddProfileInfo(corProfileInfo *A, corProfileInfo *B) ;
extern int corSimpleCopyArgs(corArgList *new_arglist, corArgList *old_arglist,
        corBindEnv *rule_env, corBindEnv *dont_rename_env,
        corTermLink *&renamed_vars);
extern int corbackpatch_bindenvs(corStackMark start, corBindEnv *dont_rename_env,
        corBindEnv *rule_env, corTermLink *&renamed_vars);

extern void corsprint_rule(corClause *rule, char *str) ;
extern int corhas_grouping_args(corArgList &args);
void cordo_dump(corTable * F, corRuleInfo * rInfo, corBindEnv * new_env,
    corArgList* new_arglist);

/*****************************************************************************/

static char rule_buf[MAX_RULE_LENGTH];


corArrayBindEnv corUnusedEnv(0);

/*****************************************************************************/

#ifdef DO_EXPLAIN
#ifdef DO_PROFILE
#define HANDLE_SUCCESS { \
    if ((rel->local_options & REL_EXPLAIN) && (corT_Stack.count > 0))\
       cordo_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
        (pinfo->i_newfacts)++ ; \
    }
#else
#define HANDLE_SUCCESS  { \
    if ((rel->local_options & REL_EXPLAIN) && (corT_Stack.count > 0)) \
       cordo_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
              }
#endif
#else
#ifdef DO_PROFILE
#define HANDLE_SUCCESS  { (pinfo->i_newfacts)++ ; }
#else
#define HANDLE_SUCCESS  { }
#endif
#endif

/****************************************************************************/
corRuleInfo::corRuleInfo(corClause *rul)
{
    num_vars = rul->num_var_names;
    num_literals = rul->num_preds;
    arg_list = new corArgList[rul->num_preds + rul->head_deletes_count + 1];
    relations = new int[rul->num_preds + rul->head_deletes_count + 2];
    offsets = new int[rul->num_preds + rul->head_deletes_count + 2];

    this->rule = rul;
    corsprint_rule(rul, rule_buf);
    char *buf = new char[strlen(rule_buf)+1];
    strcpy(buf, rule_buf);
    corProfileInfo *temp = new corProfileInfo(buf);
    pInfo = *temp;
    delete [] buf;
    delete temp;
}

/*****************************************************************************/
corRuleInfo::~corRuleInfo()
{
    delete [] arg_list;
    delete [] relations;
    delete [] offsets;
}

/*****************************************************************************/
/*
    Every time evaluate is called a number of structures need to be
    initialized. However, this is a big overhead if there is going
    to be allocation and deallocation for each evaluation of a seminaive
    rule.  So we save them instead of deallocating them, and reuse them
    on successive calls.  * We assume no parallelism however.
     TO DO:  Since space is never deallocated, we could be in trouble with
     recursive module calls.  We may need to keep a count of saved up
    structs, and free them if there are too many saved up structures.
*/

/*****************************************************************************/

#define MARKLIST(X) ((markOffsets[X]<0) ? NULL : \
            rm_array->markList[markOffsets[X]])

#define NEGATED1(index) (rInfo_relations[index] < 0)


/*****************************************************************************/
/*
    Evaluates the query in a top-down fashion. The Table structure
    F is defined in interp.h. It is a structure containing pointers to the
    various relations that will have to be accessed. It is
    initialized at query execution time with pointers to current relations
    and is a class variable of PipelinedExecInfo.The Solve function
    accesses all relations via the level of indirection provided by Table.

    The processing for a rule is done in a left-to-right manner
    The index 'current' counts how far we are along in the join.
    It represents the position of the current rhs literal being
    processed.

    We use top-down evaluation with backtracking
 
    Returns 1 on success and 0 on failure
*/

int corPipelinedExecInfo::Solve()
{
    // for efficiency
    int rInfo_num_literals = rInfo->num_literals;
    int *rInfo_relations = rInfo->relations;
    corTupleIterator *tuple_its_index;

    // Execution of a Pipelined module - top down using backtracking

    for(; current >= 0;) {
        // Exit statement from for loop. When current < 0, it implies that
        // no more facts can be generated from this rule evaluation, i.e.
        // it has backtracked past its starting point.

        // note:: it is redundant to have both index and current
        index = current;

        // Check to see if rule evaluation has been succesful.
        // This happens when a valid tuple for the last rhs literal
        // has been found. When this happens, we are going to return

        if ( current >= rInfo_num_literals ) {
            // decrement current so that the next time we begin here 
            current = rInfo_num_literals -1;
            backtrack_index = successBacktrack;
            direction = BACKWARD;
            return 1;
        }

        tuple_its_index = tuple_its[index];

        // To do a scan of each relation on the rhs, a TupleIterator is
        // needed on that relation. If this is the first time the relation
        // is reached during the evaluation, a TupleIterator needs to be created
        // for it.

        if (direction) {        // == FORWARD
            // iterator needs to be initialized
            tuple_its_index->reset();
            tuple_its_index->calling_pei = this;
            backtrack_index = index;
        }

        if (index == backtrack_index) {
            // Get the next tuple for the current relation in the nested
            // loops join.

            tuple_its[index]->get_next_tuple();

            // If there is a valid tuple, go on to the next rhs relation.
            if ( (!tuple_its_index->no_match() && !NEGATED1(index))
                 || (tuple_its_index->no_match() && NEGATED1(index)) )
            {
                         current ++;
                         direction = FORWARD;
                         continue;
            }
            else {
                backtrack_index = ((direction) ?
                    firstFailureBacktrack[current]
                    :
                    nextFailureBacktrack[current]);
            }
        }

        // On failure to find a valid tuple, backtrack to the previous
        // rhs literal.

        // Undo bindings BEFORE
        // deleting tuple_iterator (and the tuple_env in it).

        tuple_its_index->release(); // releases state of tuple iterator

        // Decrement current so that the join evaluation backtracks to the
        // previous rhs relation, and a fresh tuple for that relation will
        // be fetched in the next iteration through the for loop.
        current--;
        direction = BACKWARD;
    }

    // Have failed 
    stackmark0.pop_to();

    return 0;
}

/*****************************************************************************/
//    Dump out information to be used by the Explain tool

void cordo_dump(corTable * F, corRuleInfo * rInfo, corBindEnv * new_env,
    corArgList* new_arglist)
{
    FILE *fp = corexEnv->explain_fp;

    if (!fp) {
        return;
    }

    for (int j=0; j <= rInfo->num_literals; j++) {
        if (j==rInfo->num_literals) {
            // First Argument
            fprintf(fp, "def_%s(",
                 corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->head->predicate->name->print(new_env, fp);
            fputs("(",fp);
            ((corArgList *)new_arglist)->print_dump(new_env, fp);
            fputs(")",fp);
        }
        else {
            fprintf(fp, "use_%s(",
                corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->preds[j]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[j].print_dump(new_env, fp);
            fputs(")",fp);
        }

        fputs(", [",fp);

        rInfo->rule->head->predicate->name->print(new_env, fp);
        fputs("(",fp);
        ((corArgList *)new_arglist)->print_dump(new_env, fp);
        fputs(")",fp);

        for (int i=0; i < rInfo->num_literals; i++) {
            fputs(",",fp);
            rInfo->rule->preds[i]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[i].print_dump(new_env, fp);
            fputs(")",fp);
        }

        fputs("]).\n",fp);
    }

    //    Dump info for two new relations
    for (int j1=0; j1 <= rInfo->num_literals; j1++) {
        if (j1==rInfo->num_literals) {
            // First Argument
            fprintf(fp, "def_new_%s(",
                 corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->head->predicate->name->print(new_env, fp);
            fputs("(",fp);
            ((corArgList *)new_arglist)->print_dump(new_env, fp);
            fputs(")",fp);
        }
        else {
            fprintf(fp, "use_new_%s(",
                 corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->preds[j1]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[j1].print_dump(new_env, fp);
            fputs(")",fp);
        }

        fputs(", [",fp);

        rInfo->rule->head->predicate->name->print(new_env, fp);
        fputs("(",fp);
        ((corArgList *)new_arglist)->print_dump(new_env, fp);
        fputs(")",fp);

        for (int i1=0; i1 < rInfo->num_literals; i1++) {
            fputs(",",fp);
            rInfo->rule->preds[i1]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[i1].print_dump(new_env, fp);
            fputs(")",fp);
        }

        fputs("]).\n",fp);
    }
    fflush(fp);
}

/*****************************************************************************/
