// *************************************************************************
//    SccInfo.C
// *************************************************************************

#include "SccInfo.h"

#include "Globals.h"
#include "Literal.h"
#include "BuiltinSymbols.h"
#include "Compile.h"
#include "CompileRule.h"
#include "Execute.h"
#include "Unparse.h"

extern void corEndOfScc(corSccInfo* scc, int iters, FILE *outf);
// *************************************************************************
corSccInfo::corSccInfo(corQFMaterializedModuleData *md, corScc *scc,
    corPredicate *ext_magicpred, int id)
{
    corPredAnnotations *cur;
    int i;

    has_hidden_relations = 0;
    module = md;
    this->scc = scc;
    scc_id = id;
    this->createSNStructs(ext_magicpred);
    for (i=0; i < numPreds; i++) {
        for(cur = md->prioritize_annotations; cur; cur=cur->next()) {
            if (cur->pred==pred_ptrs[i]->name)
                has_hidden_relations = 1;
        }
    }
    rec_rule_start = 0;
    pred_success_arr = 0;
    scc->sccInfo = this; // added by Ken
}

// *************************************************************************
corSccInfo::~corSccInfo()
{
    if (numPreds) {
        delete [] pred_ptrs;
        delete [] offset_arr;
        delete [] pred_success_arr;
        delete [] sn_rules;
        delete [] sn_offset_arr;
    }
}

// *************************************************************************
int corSccInfo::allocateRuleStructures(corPredicate *ext_magicpred)
{
    // First, allocate enough memory to hold all seminaive rule structures
    numSNRules = 0;
    corPredicate *pred; corClause *clause;
    int i, j;

    for (pred = scc->preds; pred != NULL; pred = pred->next)
    {
        for (clause = pred->clauses; clause != *pred->clause_tail;
            clause = clause->pnext)
        {
            if (!clause->seen) continue;
            if (!clause->recursive) {
                if (! module->module_info.SaveModule) {
                    numSNRules++;
                    continue;
                }
                else {
                    // with SaveModules have to count the number of literals
                    // whose predicates are defined in lower SCCs.
                    // NOTE: something special needs to be done for non-magic
                    // modules
                    int thisclausesn = 0;
                    int ipred = 0;
                    for (; ipred < clause->num_preds; ipred++) {
                        // check to see if it is in a lower SCC
                        // versus being an external predicate
                        if ((clause->preds[ipred]->predicate->clauses) ||
                            (clause->preds[ipred]->predicate->coris_magic()))
                        { 
                            numSNRules++;
                            thisclausesn++;
                        }
                    }
                    if (thisclausesn == 0) { 
                        // all predicates in body are external to module
                        numSNRules++; 
                    }
                    continue;
                }
            }
            for (int ipred = 0; ipred <    clause->num_preds; ipred++) {
                // Treat negated literals as base !!
                if (clause->preds[ipred]->negated) continue;

                // The following is to special case the creation of the
                // seminaive structures for the case of Ordered Search.
                // Three special things need to be done
                // (1) `not p' is always treated as base, even if p is derived
                // (2) done_m_p is treated as derived even though there are no
                //        rules for it.
                // (3) any literal following a done_m_p is treated as base,
                //        even if p is derived. (this also covers case 1)

                if (module->module_info.UseOrdSearch) {
                    if ((ipred > 0) &&
                        (clause->preds[ipred-1]->predicate->coris_mp_done()))
                            continue;

                    // Treat mp_done literals as derived
                    if (clause->preds[ipred]->predicate->coris_mp_done())
                        clause->preds[ipred]->predicate->scc = scc;
                }
                if (clause->preds[ipred]->predicate->scc != scc) {
                    if (module->module_info.SaveModule) {
                        // check to see if it is in a lower SCC
                        // versus being an external predicate
                        if ((clause->preds[ipred]->predicate->clauses) ||
                            (clause->preds[ipred]->predicate->coris_magic())) 
                                numSNRules++;
                    }
                    continue;
                }
                numSNRules++;
            }
        }
    }

    sn_rules = new corSemiNaive[numSNRules]; 
    for (i = 0; i < numSNRules; i++)
        sn_rules[i].rule_id = i;

    sn_offset_arr = new int[numSNRules];

    // find the number of predicates in this Scc
    for (numPreds = 0, pred = scc->preds; pred != NULL;
        ++numPreds, pred = pred->next);

    // now fill in the information about lower Sccs, which is useful in
    // the case of SaveModules
    corPredicate    *ext_magicquerypred = NULL;
    if (module->numSccs == 1) { // this is the first Scc
        if (module->module_info.SaveModule) {
            ext_magicquerypred = ext_magicpred;
        }
        if (ext_magicquerypred)
            num_lowerPreds = 1;
        else
            num_lowerPreds = 0;
    }
    else { // add the numPreds and num_lowerPreds of the previous scc
        ASSERT((module->numSccs > 1));
        num_lowerPreds =
            module->sccArray[module->numSccs-2]->num_lowerPreds +
            module->sccArray[module->numSccs-2]->numPreds;
    }

    // assign the predicate pointers and the offset array
    if (!module->module_info.SaveModule) {
        pred_ptrs = new corPredicate *[numPreds];
        offset_arr = new int[numPreds];
    }
    else {
        pred_ptrs = new corPredicate * [numPreds + num_lowerPreds];
        offset_arr = new int [numPreds + num_lowerPreds];
    }
    for (i = 0, pred = scc->preds; pred != NULL; i++,pred = pred->next) {
        pred_ptrs[i] = pred;
        // allow at least one iteration
    }
    if (module->module_info.SaveModule) {
        // get the predicate pointers from the lower scc
        if (module->numSccs > 1) {
            corPredicate **lower_pred_ptrs =
                module->sccArray[module->numSccs-2]->pred_ptrs;
            ASSERT((i == numPreds));
            for (j=0; j < num_lowerPreds; j++, i++) {
                pred_ptrs[i] = lower_pred_ptrs[j];
            }
        }
        else { // module->numSccs == 1
            ASSERT((num_lowerPreds <= 1));
            if (num_lowerPreds == 1) {
                pred_ptrs[numPreds] = ext_magicquerypred;
            }
        }
    }
    return numSNRules;
}

// *************************************************************************
/*
    Routine that creates the seminaive structures in each Scc structure 
*/

void corSccInfo::createSNStructs(corPredicate *ext_magicpred)
{
    // allocate sn_rules, pred_ptrs, offset_arr
    allocateRuleStructures(ext_magicpred);

    // Now fill in values in seminaive rule structures
    corPredicate *pred; corClause *clause;

    int cur_snr = -1, i ,j;

    //    There is a correspondence between the predicate positions in 
    //    this->pred_ptrs, and the offsets in the frame given by 
    //    this->offset_arr of those predicates in rm_Arr.
    
    for (i=0; i<numPreds; i++)
        offset_arr[i] = corget_offset(module->F, pred_ptrs[i]);
    if (module->module_info.SaveModule) {
        for (i=numPreds; i < numPreds+num_lowerPreds; i++) 
            offset_arr[i] = corget_offset(module->F, pred_ptrs[i]);
    }

    //    Consider non-recursive clauses in this scc first.
    for (j=0, pred = scc->preds; pred != NULL; ++j, pred = pred->next)
    {
        for (clause = pred->clauses; clause != *pred->clause_tail;
            clause = clause->pnext)
        {
            if (!clause->seen) continue;     
            if (clause->recursive) continue;

            // we know it is a non-recursive clause by now
            int thisclausesn = 0; 
            if (module->module_info.SaveModule) { 
                // have to generate SN rules corresponding to lower predicates
                int ipred, jpred;
                for (ipred=0; ipred < clause->num_preds; ipred++) {
                    // all body literals are in lower sccs or external
                    // predicates
                    ASSERT((clause->preds[ipred]->predicate->scc != scc));
                    if ((clause->preds[ipred]->predicate->clauses) ||
                        (clause->preds[ipred]->predicate->coris_magic()))
                    { 
                        // lower scc or containing MagicQueryPredKind without
                        // any rules

                        thisclausesn++; // count number of sn rules generated 
                        cur_snr++; 
                        ASSERT((cur_snr < numSNRules));
                        sn_rules[cur_snr].first = ipred;
                        sn_rules[cur_snr].recursive = 0;
                        sn_rules[cur_snr].scc = this;
                        sn_offset_arr[cur_snr] = j;

                        sn_rules[cur_snr].Add_Rule_Info(
                            module->ruleArray->rInfoPtrs[clause->rule_number],
                            clause, module->module_info.NonGroundFacts);
                        for (jpred = 0; jpred < clause->num_preds; jpred++) {
                            corLiteral *bodypred = clause->preds[jpred];

                            if (ipred == jpred) {
                                // NOTE: check whether external predicates have 
                                // such rmarks defined on it or not!
                                // they may be needed 
                                sn_rules[cur_snr].markOffsets[2*jpred] =
                                    corFind_RMark_Offset(this,bodypred->predicate);
                            }
                            else {
                                sn_rules[cur_snr].markOffsets[2*jpred] = -1;
                            }
                
                            if (bodypred->predicate->clauses ||
                                bodypred->predicate->coris_magic())
                            {
                                int tmp;
                                
                                tmp = corFind_RMark_Offset(this,
                                    bodypred->predicate);
                                if (ipred    > jpred) {
                                    sn_rules[cur_snr].markOffsets[2*jpred+1] =
                                        tmp;
                                }
                                else {
                                    sn_rules[cur_snr].markOffsets[2*jpred+1] =
                                        tmp+1;
                                }
                            }
                            else {
                                sn_rules[cur_snr].markOffsets[2*jpred+1] = -1;
                            }
                        }
                    }
                }
            }

            if (thisclausesn ==0) { 
                // Either no SaveModule or a rule with only external predicates
                // in a SaveModule
                // Same actions need to be performed
                cur_snr++;
                ASSERT((cur_snr < numSNRules));
                sn_offset_arr[cur_snr] = j;
                sn_rules[cur_snr].scc = this;

                // the following function encapsulates the initialization of
                // a non-recursive seminaive rule. This is done so that 
                // imperative rules from the coral prompt can also be
                // initialized using the same code.
                corcompile_nonrec_rule(sn_rules[cur_snr], clause,
                    module->ruleArray->rInfoPtrs[clause->rule_number],
                    module->module_info.NonGroundFacts);
            }
        }
    }
    
    // Now handle recursive rules

    if (scc->recursive) { // If there are recursive rules:
        if (module->module_info.SaveModule) {
            // have to generate non-recursive SemiNaive rules from
            // recursive rules
            for (j=0, pred = scc->preds; pred != NULL; ++j, pred = pred->next)
            {
                for (clause = pred->clauses; clause != *pred->clause_tail;
                    clause = clause->pnext)
                {
                    if (!clause->seen) continue;
                    if (!clause->recursive) continue;

                    int ipred, jpred, i;
                    for (ipred=0; ipred < clause->num_preds; ipred++) {
                        if (module->module_info.UseOrdSearch) {
                            if ((ipred > 0) &&
                              (clause->preds[ipred-1]->predicate->coris_mp_done()))
                                continue;
                        }

                        // disregard literals defined in the same scc
                        if (clause->preds[ipred]->predicate->scc == scc)
                            continue;
                
                        if ((clause->preds[ipred]->predicate->clauses) ||
                            (clause->preds[ipred]->predicate->coris_magic()))
                        {
                            // lower scc or containing MagicQueryPredKind
                            // without any rules

                            cur_snr++;
                            ASSERT((cur_snr < numSNRules));
                            i = corhas_fail_in_body(clause);
                            if (i >= 0) {
                                sn_rules[cur_snr].first = i;
                            }
                            else {
                                sn_rules[cur_snr].first = ipred;
                            }

                            sn_rules[cur_snr].recursive = 0;
                            sn_rules[cur_snr].scc = this;
                            sn_offset_arr[cur_snr] = j;
                            sn_rules[cur_snr].Add_Rule_Info(module->ruleArray->
                                rInfoPtrs[clause->rule_number],clause,
                                module->module_info.NonGroundFacts);

                            for (jpred=0;jpred < clause->num_preds; jpred++) {
                                corLiteral *bodypred = clause->preds[jpred];

                                if (ipred == jpred) {
                                    sn_rules[cur_snr].markOffsets[2*jpred] =
                                        corFind_RMark_Offset(this,
                                        bodypred->predicate);
                                }
                                else {
                                    sn_rules[cur_snr].markOffsets[2*jpred] = -1;
                                }

                                if (bodypred->predicate->scc == scc) {
                                    // only until the global old mark; not the
                                    // whole relation
                                    sn_rules[cur_snr].markOffsets[2*jpred+1] =
                                        corFind_RMark_Offset(this,
                                        bodypred->predicate);
                                }
                                else if (bodypred->predicate->clauses ||
                                    bodypred->predicate->coris_magic())
                                {
                                    int tmp = corFind_RMark_Offset(this,
                                        bodypred->predicate);
                                    if (ipred > jpred)
                                        sn_rules[cur_snr].markOffsets[2*jpred+1]
                                            = tmp;
                                    else
                                        sn_rules[cur_snr].markOffsets[2*jpred+1]
                                            = tmp+1;
                                }
                                else { // external non querymagic predicate 
                                    sn_rules[cur_snr].markOffsets[2*jpred+1] = -1;
                                }
                            } 
                        }
                    }
                }
            }
        }

        // now generate recursive SemiNaive rules from recursive rules
        for (j=0, pred = scc->preds; pred != NULL; j++, pred = pred->next)
        {
            for (clause = pred->clauses; clause != *pred->clause_tail;
                clause = clause->pnext)
            {
                if (!clause->seen) continue;
                if (!clause->recursive) continue;

                int ipred, jpred, i;
                for (ipred = 0; ipred < clause->num_preds; ipred++) {
                    //
                    //    Treat negated literals as base !!
                    //
                    if (clause->preds[ipred]->negated) continue;

                    // with Ordered Search, have to special case the 
                    // negated literals as well as the grouped literal.
                    // the check below assumes that both follow a 
                    // done_m_p literal.

                    if (module->module_info.UseOrdSearch) {
                        if ((ipred > 0) &&
                            (clause->preds[ipred-1]->predicate->coris_mp_done()))
                                continue;
                    }

                    // don't need to special case for mp_done since it is
                    // covered by the next if condition

                    if (clause->preds[ipred]->predicate->scc != scc)
                        continue;

                    cur_snr++; 
                    ASSERT((cur_snr < numSNRules));
                    i = corhas_fail_in_body(clause);
                    if (i >= 0) {
                        sn_rules[cur_snr].first = i;
                    }
                    else {
                        sn_rules[cur_snr].first = ipred;
                    }            
                    sn_rules[cur_snr].recursive = 1;
                    sn_rules[cur_snr].scc = this;
                    sn_offset_arr[cur_snr] = j;
                    sn_rules[cur_snr].Add_Rule_Info(
                            module->ruleArray->rInfoPtrs[clause->rule_number],
                            clause, module->module_info.NonGroundFacts);
                    
                    for (jpred = 0; jpred < clause->num_preds; jpred++) {
                        corLiteral *bodypred = clause->preds[jpred];
                        
                        if (ipred == jpred) {
                            sn_rules[cur_snr].markOffsets[2*jpred] = 
                                corFind_RMark_Offset(this, bodypred->predicate);
                        }
                        else {
                            sn_rules[cur_snr].markOffsets[2*jpred] = -1;
                        }
                        
                        if (bodypred->predicate->scc != scc)
                            sn_rules[cur_snr].markOffsets[2*jpred+1] = -1;
                        else {
                            int tmp = corFind_RMark_Offset(this,
                                bodypred->predicate);
                            if (ipred > jpred)
                                sn_rules[cur_snr].markOffsets[2*jpred+1] =tmp;
                            else
                                sn_rules[cur_snr].markOffsets[2*jpred+1] =tmp+1;
                        }
                    }
                }
            }
        }
    }
    // when sn_rule generation is done, add labels to them :added by Ken
    for (i = 0; i<=cur_snr; i++)
            sn_rules[i].Add_label();

}

// *************************************************************************
void corSccInfo::display_annotation_info(corTable *F) 
{
    FILE *file = corexEnv->trace_file;
    corPredAnnotations *cur;
    corQFModuleData *md = F->moduledata;
    
    for(int i=0; i < numPreds; i++) {
        Name predname = F->rels[offset_arr[i]].relPtr->name;
        
        for(cur= md->agg_sel_annotations; cur; cur=cur->next()) {
            if (cur->pred==predname) {
                fprintf(file,"%s%s    ", "@",corSymbolString(corAggSelectionSymbol));
                cur->print(file);
                fprintf(file, ".\n");
            }
        }
        for(cur= md->prioritize_annotations; cur; cur=cur->next()) {
            if (cur->pred==predname) {
                fprintf(file,"%s%s    ", "@",corSymbolString(corPrioritizeSymbol));
                cur->print(file);
                fprintf(file, ".\n");
            }
        }
        fprintf(corexEnv->trace_file, "%s: subsumption checking - %s \n\n",
            F->rels[offset_arr[i]].relPtr->name->string(),
            F->rels[offset_arr[i]].relPtr->check_subsum? "on" : "off");
        /* PRAVEEN:: Add the index information dumping routines here. */
    }
    fprintf(file,"---------------------------------------\n\n");
}

// *************************************************************************
corRMarkArray * corSccInfo::setupSolve(corTable *F)
{
    int sn_count;
    int i;
    
    // Create array of RMarks
    // By default, all RMark *s are NULL 
    corRMarkArray *rm_Arr = NULL;
    corRMarkArray *hidden_rel_rmarks = NULL;

    rm_Arr = F->module_env->scc_envs[this->scc_id]->rm_Arr;
    if (module->module_info.SaveModule) {
        // set up RMarks for the case of non-recursive SN rules for SaveModules
        corStorageRelation *hrel = NULL;
        for (i=0; i < numPreds+num_lowerPreds; i++) {
            hrel = (corStorageRelation*)(F->rels[offset_arr[i]].relPtr);
            if (hrel) {
                ASSERT((hrel->there_are_marks()));
                // set up the old marks to the status of the relations as of
                // the end of the previous call to the SaveModule
                rm_Arr->markList[2*i] = hrel->headMark.prev;
            
                // we don't need to get new marks for any of the relations 
                // for SaveModule
            }
        }
    }

    if (has_hidden_relations) {
        hidden_rel_rmarks = new corRMarkArray(2*numPreds);
        for (i=0; i < numPreds; i++) {
            if (F->rels[offset_arr[i]].relPtr->priority_info)
                hidden_rel_rmarks->markList[2*i+1] =
                    F->rels[offset_arr[i]].relPtr->getMark();
        }
    }
    
    if ((DISPLAY_RULES(module)) && (!sn_rules[0].recursive) &&
        (!corexEnv->C_quiet_mode_default))
            corPrintSccNonRecursiveRules(this, corexEnv->trace_file);
    
    //    For Basic SemiNaive, evaluate non-recursive rules
    //    For Predicate SemNaive, this is incorporated into the
    //        first iteration of the recursive rules.

#define CUR_POS sn_offset_arr[sn_count]
    for (sn_count = 0; sn_count < numSNRules; sn_count++) {
        if (sn_rules[sn_count].recursive)
            break;
        else if (!module->use_psn) {
            //if (!sn_rules[sn_count].ignore_this_rule) {
                F->moduledata->curr_rInfo = sn_rules[sn_count].rInfo;
                sn_rules[sn_count].evaluate(F, rm_Arr);
            //}
        }
    }
#undef CUR_POS


    // I dont know if it is possible that the lowest scc has a rule for
    // the m_query relation , but the rule is non-recursive. But just in
    // case this is possible, i catch it here, and insert the m_query
    // seed tuple at this point
    if (corMagicRelation && module->use_psn && (sn_count == numSNRules)) {
        corMagicRelation->insert_new(*corMagicArgs, corMagicBindEnv);
        corMagicRelation = NULL; corMagicArgs = NULL; corMagicBindEnv = NULL;
    }

    //    NOTE : if the scc has only non-recursive rules, evaluate them all
    //    even for PSN.

    if (module->use_psn && (sn_count == numSNRules)) {
        for (sn_count = 0; sn_count < numSNRules; sn_count++) {
            //if (!sn_rules[sn_count].ignore_this_rule) {
                F->moduledata->curr_rInfo = sn_rules[sn_count].rInfo;
                sn_rules[sn_count].evaluate(F, rm_Arr);
            //}
        }
    }

    rec_rule_start = sn_count;

    if (has_hidden_relations) {
        for (i=0; i < numPreds; i++) {
            if (F->rels[offset_arr[i]].relPtr->priority_info) {
                F->rels[offset_arr[i]].relPtr->
                    freeMark(hidden_rel_rmarks->markList[2*i+1],1);
            }
        }
    }


    if (sn_count == numSNRules) return rm_Arr;


    // Note on RMarks for recursive rules. For each predicate, an 'old' RMark
    // and a 'new' RMark are created at consecutive positions in the RMarkArray.
    // The order in which these RMarks are created defines their positions in
    // the array, and this fact is used by Find_RMark_Offset() .

    if (DISPLAY_RULES(module) && !corexEnv->C_quiet_mode_default)
        corPrintSccRecursiveRules(this, rec_rule_start, corexEnv->trace_file);

    return rm_Arr;
}

// *************************************************************************
static int evaluateNonRecursive(corSccInfo *scc, corTable *F,corRMarkArray *rm_Arr,
    int curpos)
{
    int sn_count;
    int new_tuples = 0;

    if (corMagicRelation &&
        (F->rels[scc->offset_arr[curpos]].relPtr == corMagicRelation))
    {
        corMagicRelation->insert_new(*corMagicArgs,corMagicBindEnv);
        corMagicRelation = NULL;
        corMagicArgs = NULL;
        corMagicBindEnv = NULL;
        new_tuples = 1;
    }

            
    for (sn_count = 0; sn_count < scc->rec_rule_start; sn_count++) {
        if (scc->sn_offset_arr[sn_count] == curpos) {
            //if (!scc->sn_rules[sn_count].ignore_this_rule) {
                F->moduledata->curr_rInfo = scc->sn_rules[sn_count].rInfo;
                new_tuples += scc->sn_rules[sn_count].evaluate(F, rm_Arr);
            // }
        }
    }

    return new_tuples;
}

// *************************************************************************
/*
    Performs one iteration over the recursive rules and returns 1 if the
    iteration should continue, and 0 if the iterative loop should be exited.
*/

int corSccInfo::iterateRecursiveRules(corTable *F, corRMarkArray *rm_Arr,
    int loop_iter_num, int iteration_number)
{
    int new_tuples = 0;
    int sn_count;
    
    if (module->use_psn) {
        // Predicate SemiNaive

        int curpos = sn_offset_arr[rec_rule_start];
        rm_Arr->markList[2*curpos+1] =
            F->rels[offset_arr[curpos]].relPtr->getMark();

        for (sn_count = rec_rule_start; sn_count < numSNRules; sn_count++) {
            //if (!sn_rules[sn_count].ignore_this_rule) {
                F->moduledata->curr_rInfo = sn_rules[sn_count].rInfo;
                new_tuples += sn_rules[sn_count].evaluate(F, rm_Arr);
            // }

            if ((sn_count == numSNRules -1) ||
                (sn_offset_arr[sn_count+1] != curpos))
            {
                int hide_em = 0;
                if (module->use_psn && (loop_iter_num == 1) &&
                    (iteration_number == 1))
                {
                    new_tuples += evaluateNonRecursive(this, F, rm_Arr, curpos);
                }

                if ((! module->module_info.SaveModule) || (loop_iter_num > 1))
                {
                    F->rels[offset_arr[curpos]].relPtr->
                        freeMark(rm_Arr->markList[2*curpos]);
                }

                if (has_hidden_relations &&
                    F->rels[offset_arr[curpos]].relPtr->priority_info)
                {
                    hide_em = 1;
                    F->rels[offset_arr[curpos]].relPtr->
                        freeMark(rm_Arr->markList[2*curpos+1], hide_em);
                    rm_Arr->markList[2*curpos+1] =
                        F->rels[offset_arr[curpos]].relPtr->getMark();
                }

                rm_Arr->New_To_Old(2*curpos);

                if (sn_count < numSNRules -1) {
                    curpos = sn_offset_arr[sn_count+1];
                    rm_Arr->markList[2*curpos+1] = F->rels[offset_arr[curpos]].
                        relPtr->getMark();
                }
            }
        }
    }
    else {
        // Basic SemiNaive
        int i, j;
        for (i=0, j =1; i < numPreds; i++, j += 2) {
            rm_Arr->markList[j] =
                F->rels[offset_arr[i]].relPtr->getMark();
        }
        
        for (sn_count = rec_rule_start; sn_count < numSNRules; sn_count++) {
            corSemiNaive *cur_sn_rule = &(sn_rules[sn_count]);
            //if (!cur_sn_rule->ignore_this_rule) {
                 F->moduledata->curr_rInfo = cur_sn_rule->rInfo;
                 new_tuples += cur_sn_rule->evaluate(F, rm_Arr);
            //}
        }


        if (has_hidden_relations) {
            for (i=0; i < numPreds; i++) {
                if (F->rels[offset_arr[i]].relPtr->priority_info) {
                    F->rels[offset_arr[i]].relPtr->
                        freeMark(rm_Arr->markList[2*i+1], /* hide_em */ 1);
                    rm_Arr->markList[2*i+1] =
                        F->rels[offset_arr[i]].relPtr->getMark();
                 }
            }
        }

        if ((! module->module_info.SaveModule) || (loop_iter_num > 1)) {
            for (i=0, j = 0; i < numPreds; i++, j+=2){
                // the old marks should not be freed, if we are in the
                // first iteration of SaveModules since these correspond to 
                // the global old pointers obtained at the end of evaluation of
                // the previous call to the module.
                F->rels[offset_arr[i]].relPtr->freeMark(rm_Arr->markList[j]);
            }
        }

        int numpreds2 = 2*numPreds;

        for (i=0, j=1; i < numpreds2; i+=2, j+=2) {
            // rm_Arr->New_To_Old(2*i);
            rm_Arr->markList[i] = rm_Arr->markList[j];
            rm_Arr->markList[j] = NULL;
         }
    }
    
    corGoalNode *g_node; 

    if (new_tuples)
        return 1;
    else {
        //    There are a bunch of special cases in which the number of
        //    new_tuples is == 0, but there is still some computation to go.
        //    However, if the value of new_tuples is non_zero, then we can
        //    return right away

        if (has_hidden_relations) {
            for (int i =0; i < numPreds; i++) {
                if (F->rels[offset_arr[i]].relPtr->priority_info) {
                    if (((corStorageRelation*)(F->rels[offset_arr[i]].relPtr))->
                        pop_from_hidden_relation() == 0)
                            continue;
                    else {
                        return 1;
                        // return, so newly unhidden tuple
                        // can be displayed to user if required.
                        // Tuple will get used in a later call on this routine
                    }
                }
            }
        }

        if (!(module->module_info.UseOrdSearch))
            return 0;

        // Ordered Search
        // have reached a fixpoint; need to get unmarked tuple 
        // from context, and put it in the appropriate relation

        int any_new_fact_in_done = 0;
        g_node = F->context->get_unmarked_goal(any_new_fact_in_done);
        
        // exit condition from Ordered Search
        if (g_node == NULL) {
            if (!any_new_fact_in_done)
                return 0;
            else
                return 2;
        }

        if ((!module->module_info.SaveModule) ||
            (! g_node->owner_rel->check_subsum) ||
            (! g_node->owner_rel->is_subsumed(g_node)))
        {
            g_node->owner_rel->insert_tuple(g_node);
            // NOTE: should remove g_node from 
            // g_node->owner_rel->for_ordsearch's index
            return 2;
        }
        return 0;
    }
}

// *************************************************************************
/*
    Iterator will be non-null ONLY if this is a top-level scc and
    staggered evaluation is desired on it.
*/

void corSccInfo::Solve(corTable *F, corTupleIterator *iterator)
{
    corRMarkArray *rm_Arr;
    int i;
    
    // if staggered evaluation is not being performed, or this is the
    // not the top level scc, or this is the first time the top level 
    // scc is being executed

    if (!(iterator) || !(iterator->state->rm_Arr)) {
        rm_Arr = setupSolve(F);

        if (iterator) {
            // first time this top-level scc is being executed
            iterator->state->rm_Arr = rm_Arr;
        }
    }
    else {
        rm_Arr = iterator->state->rm_Arr;
    }

    // Sudarshan:  Handling prioritize at the level of SN updates is critical.
    // I tried to handle hiding facts at the level of fact insertion, but
    // that caused major problems if a relation has aggregation
    // (even if non-recursive) as well as prioritize.    The trouble is that
    // if facts are hidden during aggregation, the result of aggregation is
    // garbage.    So now facts are hidden only when they are moved out of the
    // delta relations.
    //
    // The priority code works as follows.
    // a) get rmarks on prioritized relations even before exit rules,
    // b) after each iteration or exit rule firing, move tuples from
    //    deltas to the hidden relation, instead of to the previous
    //    delta relation, if the relation uses priority.

    if (rec_rule_start == numSNRules) {
        /*** if (rm_Arr) delete rm_Arr;        ****/
        if (rm_Arr) {
            for (i = 0; i< rm_Arr->array_size; i++) {
                rm_Arr->markList[i] = NULL;
            }
        }

        if (iterator) iterator->state->rm_Arr = NULL;
        if (has_hidden_relations) {
            // Shouldn't happen normally, since the
            // annotation is meaningless in this case.
            // To be safe, unhide all hidden facts.
            clean_hidden_relations(F);
        }
#        ifdef DO_PROFILE
            if ((corexEnv->GlobalRelationOptions | module->rel_options) &
                REL_PROFILE_STATS)
                    corEndOfScc(this, 0, corexEnv->trace_file);
#        endif
        return;
    }

    // loop_iter_num is useful for SaveModules since the first time
    // around, the old RMarks correspond to global old pointers and 
    // should not be deleted

    int loop_iter_num = 0;
    int ret_val = 1;

    // Evaluate recursive rules
    if (iterator) {
        // if staggered evaluation is being used, then the iter number is
        // actually stored in the iterator's state, and not in the variable
        // loop_iter_num.
        for(;ret_val;) {
            if ((ret_val = iterateRecursiveRules(F, rm_Arr, ++loop_iter_num,
                iterator->state->iter_number)) && ret_val != 2)
                    return;
        }
    }
    else {
        while (ret_val = iterateRecursiveRules(F, rm_Arr, ++loop_iter_num,
            loop_iter_num));
    }
    
    // since the loop is going through at least once, these marks 
    // will not be the global old marks 

    for (i=0; i < numPreds; i++) {
        F->rels[offset_arr[i]].relPtr->freeMark(rm_Arr->markList[2*i]);
    }

    for (i = 0; i< rm_Arr->array_size; i++) {
        rm_Arr->markList[i] = NULL;
    }

    if (iterator) iterator->state->rm_Arr = NULL;

#    ifdef DO_PROFILE
        if ((corexEnv->GlobalRelationOptions | module->rel_options) &
            REL_PROFILE_STATS)
                corEndOfScc(this, loop_iter_num, corexEnv->trace_file);
#    endif
}

// *************************************************************************
void corSccInfo::clean_hidden_relations(corTable *F) 
{
    for (int i=0; i < numPreds; i++) {
        if(has_hidden_relations    &&
            F->rels[offset_arr[i]].relPtr->priority_info)
        {
             ((corStorageRelation*)(F->rels[offset_arr[i]].relPtr))->
                pop_all_from_hidden_relation();
        }
    }
}

// *************************************************************************
void corSccInfo::print(FILE *fd, int print_preds, int print_text, 
            int print_snrules) // added by Ken
{
  fprintf(fd, "Scc #%d: \n", scc_id);
  scc->print(fd, print_preds, print_text);
  if (print_snrules) {
    for (int i=0; i<numSNRules; i++)
      sn_rules[i].print(fd);
  }
}
// *************************************************************************
