
//**************************************************************************
//    Connection.C
//
//    The Connection Class - Used to hold all information about a 
//    connection. This includes the socket being used, a ParseStack, and
//    an array of users.
//
//**************************************************************************

#include "Connection.h"
#include "Database.h"
#include "Globals.h"
#include "Assert.h"

#include "Suspension.h"

//**************************************************************************
// Connection Constructor -
//    Clears out the user array, accepts a connection on the server socket
//    (which generates a socket to talk to the client with), re-opens the
//    socket with file pointers, and puts the incoming one on the parse
//    stack.  This function also generates a default user "nobody" so
//    the client has an execution envirornment before they login(). 
//
//**************************************************************************
corConnection::corConnection(CoralSock *serv_sock, int put_on_stack) 
{
    FILE *in = NULL, *out = NULL;

    num_users = 0;
    nxt_user = 1;           // Since 0 is "nobody"
    cur_user = NOBODY;      // "nobody" is the current user

    for (int i = 0; i < MAX_USRS_PER_CONN; i++)
        userArray[i] = (corUser *) NULL;

    conn_sock = serv_sock->AcceptConnection();

    if (put_on_stack){
        in  = conn_sock->InFilePtr();
        out = conn_sock->OutFilePtr();

          // Put the connection onto the parse stack
        pStack.push(in);
    }

      // Generate user "nobody" (id is always 0)
    if (in)
        userArray[NOBODY] = new corUser(in,out,out);
    else
        userArray[NOBODY] = new corUser(stdin,stdout,stderr);

    strcpy((userArray[NOBODY])->login_name,"nobody");

      // No path to start with
    cwd[0] = '\0';
}

//**************************************************************************
// Connection Destructor -
//    Deletes all users in the user array, and deletes default user
//    "nobody". Closes the socket to the client.
//
//**************************************************************************
corConnection::~corConnection()
{
    for (int i=1; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i])
            DeleteUser(i);

      // Now get rid of nobody
    corsuspension.killall(conn_id, 0);
    delete userArray[0];

    if (conn_sock)
        delete conn_sock;
}

//**************************************************************************
// NewUser() -
//    Generates a new user using the name and password given, and sets
//    them up. If this new user is the only user on this connection,
//    (excluding "nobody"), the current user is switched to them. 
//
//    Return: The user_id of the newly generated user, or < 0 if an
//        error occurs.
//
//**************************************************************************
int corConnection::NewUser(char *name, char *passwd)
{
    if (!name || !(*name))
        return(-1);

      // user "nobody" is not counted in num_users
    if (num_users >= (MAX_USRS_PER_CONN-1))
        return(-1);

      // Do not allow user nobody to be created in this fashion
    if (strcmp(name,"nobody") == 0)
        return(-1);

      // NYI - Check that this is this users correct password


    while(userArray[nxt_user])
        nxt_user = (nxt_user + 1) % MAX_USRS_PER_CONN;

    FILE *in  = conn_sock->InFilePtr();
    FILE *out = conn_sock->OutFilePtr();

    userArray[nxt_user] = new corUser(in,out,out);
    strncpy((userArray[nxt_user])->login_name, name, 9);
    strncpy((userArray[nxt_user])->password, passwd, 9);

    (userArray[nxt_user])->user_id  = nxt_user;

      // This is added simply as a convienence, if you are nobody
      // (like when you first connect) login() also switches you
      // to be the new user you just created.
    if (cur_user == NOBODY)
        useUser(nxt_user);

    num_users++;
    return(nxt_user);
}

//**************************************************************************
// DeleteUser() -
//    Deletes the user specified by user_id, freeing up the space in the
//    user array and deleting any suspended queries they have.
//    If they are the current user, the current user becomes the default
//    user "nobody".  The user "nobody" cannot be deleted with this
//    function, they are only deleted when the connection is closed.
//
//**************************************************************************
void corConnection::DeleteUser(int user_id)
{
    if ((! user_id) || (! userArray[user_id]))
        return;

      // delete suspended queries for this user
    corsuspension.killall(conn_id, user_id);

    delete userArray[user_id];
    userArray[user_id] = (corUser *) NULL;

    nxt_user = user_id;

    if (user_id == cur_user){
      // Logging self out - become "nobody"
      useUser(NOBODY);
    }

    num_users--;
}

//**************************************************************************
// useConnection() -
//    Switch all global pointers to make this connection the current
//    connection. This function also switches the global current user
//    to be the current user of this connection. 
//
//    Return: 1 on success, 0 if current user of this connection didn't
//        exist.
//
//**************************************************************************
int corConnection::useConnection()
{
    corcurConn = this;

      // Set the global parse stack to this one
    corparseStack = &pStack;
    if (! corparseStack->is_empty() )
        corparseStack->reset_position();

    if (!useUser(cur_user)){
        // The user specified doesn't exist - become "nobody"
        useUser(NOBODY);
        return(0);
    }
    
    return(1);
}

//**************************************************************************
// useUser() -
//    Switch all global pointers to make this user the current
//    user.
//
//    Return: 1 on success, 0 if the user didn't exist.
//
//**************************************************************************
int corConnection::useUser(int new_user)
{
    if (! userArray[new_user])
        return(0);

    corcurUser = userArray[new_user];
    cur_user = new_user;

      // What else needs to be switched ??
    corexEnv = &(corcurUser->corexEnv);
    corCurDB = corcurUser->corCurDB;

    return(1);
}

//**************************************************************************
// FindUser() -
//    Finds the user given by name. If the name is not unique, returns
//    the first user that matches.
//
//    Return: The user_id of the user with this name, or < 0 if no user
//        by that name is on this connection. 
//
//**************************************************************************
int corConnection::FindUser(char *name)
{
    for (int i=0; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i] && (strcmp(name,userArray[i]->login_name) == 0))
            return(i);

    return(-1);
}

//**************************************************************************
// ListWho()
//    Lists all users for this connection, printing the information to
//    the file pointer specified. The default user "nobody" is ignored.
//
//    Return: The number of users found.
//
//**************************************************************************
int corConnection::ListWho(FILE *fp)
{
    int cnt=0;

    fprintf(fp,"Connection %d:\n", conn_id);
    for (int i=1; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i]){
            cnt++;
            fprintf(fp,"%4d   %s\n", i, userArray[i]->login_name);
        }

    return(cnt);
}

//**************************************************************************
