//**************************************************************************
//    FullMagic.C
//**************************************************************************
 
#include "FullMagic.h"

#include "SupMagic.h"

#include "RelationMisc.h"
#include "Literal.h"
#include "Globals.h"
#include "Unparse.h"
#include "SipOrder.h"

//**************************************************************************
//    Assumptions:  Adorned rules (in the structure as_rules) have been sorted
//    in sip order already.
//**************************************************************************

int opt_share_functors = 0;

static int debugmagic = 0;

//    If set, suppress generation of magic literals for predicates with
//    all arguments free.
static int suppress_allfree = 0;

//**************************************************************************
void nomagic(struct corcollection *rules)
{
    int j;
    corClause *magicrule;
    extern struct corcollection magic_rules[1];    // Main.C

    corinit_collection(magic_rules, 0);
    FOR_EACH_RULE(rule, rules) {
        magicrule = corcopy_rule(rule);
        magicrule->head = rule->head;
        coralloc_preds(magicrule, rule->num_preds );
        for (j=0; j<rule->num_preds; ++j)
            magicrule->preds[j] = rule->preds[j];
        coradd_rule(magicrule, magic_rules);
    } END_EACH_RULE
}

//**************************************************************************
void fullmagic(struct corcollection *ad_rules, struct corcollection *magic_rules)
{
    FOR_EACH_RULE(adrule, ad_rules) {
        //    Construct the magic rules
                if (0 && corparserStruct.CurModule.NestedSccEval && adrule->supp)
          continue; // added by Ken

        do_magic_on_rule(adrule, magic_rules);
    } END_EACH_RULE
}

//**************************************************************************
/*
    This is used to compute the number of "done_m_" literals to be added
    in the magic rule.  There are two cases:
        (1) If there is aggregation in the head, then pre-processing
            guarantees that only one "done_m_" literal needs to be 
            added to the rule.
        (2) If there is genuine negation in the body of the rule, add
            as many "done_m_" as there are negated literals.
*/

int compute_deriv_negated(corClause *adrule)
{
    int ret_val = 0;

    //    For monotonic programs, there is no need to add any done_m_ literals.
    if (corparserStruct.CurModule.Monotonic)
        return ret_val;

    if (corhas_grouping_args(adrule->head->args))
        ret_val = 1;
    else {
        //    There is genuine negation in the head of the rule
        for (int i=0; i < adrule->num_preds; ++i) {
            if (adrule->preds[i]->negated && (!adrule->preds[i]->base_pred)) {
                // negated and derived predicates
                ++ret_val;
            }
        }
    }
    return ret_val;
}

//**************************************************************************
void do_magic_on_rule(corClause * adrule, struct corcollection * magic_rules)
{
    int i,j;
    corClause *magicrule;

    if (debugmagic) {
        printf("Original rule :\n") ;
        corprint_rule(adrule,stderr);
    }

    //    NOTE: The following code is for Ordered Search.
    //    It finds the number of derived negated literals in the body of 
    //    a rule and hence the number of "done_m_" literals to be added
    //    Also, it always uses ordered search whenever a derived negated
    //    literal is encountered.

    int num_deriv_negated = compute_deriv_negated(adrule);
    // number of negated derived literals in adrule

    for (i=0; i < adrule->num_preds; ++i)
        create_derived_magic_rules(adrule, magic_rules, i);

    //    Construct the modified rule
    magicrule = corcopy_rule(adrule);

    if (! (allargsfree(& adrule->head->adorn) && suppress_allfree)) {
        int num_bound_functor_args = 0;
        if (opt_share_functors) {
            //    Optimization to let the head share functors with the magic
            //    predicate in the body
            num_bound_functor_args = count_bound_functor_args(adrule->head);

            //    Create new variables for functor args
            make_new_rule_vars(magicrule, num_bound_functor_args);
        }
        coralloc_preds(magicrule, adrule->num_preds + 1 + num_deriv_negated +
            num_bound_functor_args);
        magicrule->head = build_head_pred(magicrule, adrule->head);
        magicrule->preds[0] = build_magic_pred(magicrule, adrule->head,
            opt_share_functors);
        if (opt_share_functors) {
            create_functor_literals(magicrule);
        }
        for (j=num_bound_functor_args+1,i=0; i<adrule->num_preds; ++j, ++i) {
            // NOTE: The following code is for Ordered Search
            if ((corparserStruct.CurModule.UseOrdSearch) &&
                (!corparserStruct.CurModule.Monotonic) &&
                ((corhas_grouping_args(adrule->head->args)) ||
                (adrule->preds[i]->negated)) &&
                (!adrule->preds[i]->base_pred))
            {
                magicrule->preds[j] = build_magic_pred(magicrule,
                    adrule->preds[i], 0) ;
                magicrule->preds[j++]->kind = COR_DONE_MAGIC;
            }
            magicrule->preds[j] = adrule->preds[i];
         }
    }
    else {
        //    Copy out the original rule
        magicrule->head = adrule->head;
        coralloc_preds(magicrule, adrule->num_preds + num_deriv_negated);
        for (j=0, i=0; i<adrule->num_preds; ++j,++i) {
            // NOTE: The following code is for Ordered Search
            if ((corparserStruct.CurModule.UseOrdSearch) &&
                (!corparserStruct.CurModule.Monotonic) &&
                ((corhas_grouping_args(adrule->head->args)) ||
                (adrule->preds[i]->negated)) &&
                (!adrule->preds[i]->base_pred))
            {
                magicrule->preds[j] = build_magic_pred(magicrule,
                    adrule->preds[i],0) ;
                magicrule->preds[j++]->kind = COR_DONE_MAGIC;
            }
            magicrule->preds[j] = adrule->preds[i];
        }
    }

    magicrule->head_deletes_count = adrule->head_deletes_count;
    magicrule->head_deletes = adrule->head_deletes;
    coradd_rule(magicrule, magic_rules);
}

//**************************************************************************
/*
    Generate a magic rule for each derived right-hand side predicate that
    has an incoming arc in the sip.   Currently handles only linear sips.
*/

void create_derived_magic_rules(corClause *adrule,
    struct corcollection *magic_rules, int pred_index)
{
    int j,p;
    corClause *magicrule;
    struct corsip *s;

    if (adrule->preds[pred_index]->base_pred) return;

    if (adrule->sips->sip_list)
        s = findadsip(adrule, pred_index);
    else
        s = NULL;
    
    //    Copy proto header
    magicrule = corcopy_rule(adrule);
    magicrule->head = build_magic_pred(adrule, adrule->preds[pred_index], 0);
    if (s) {
        coralloc_preds(magicrule, numpredsinsip(&s->tail));
        for(j=0,p=firstpredinsip(&s->tail); p >= 0;
            p = nextpredinsip(&s->tail,p) )
        {
            if (p==0) {
                //    Head predicate
                if (!(allargsfree(& adrule->head->adorn) && suppress_allfree)) {
                    magicrule->preds[j++] =
                        build_magic_pred(adrule, adrule->head, 0);
                }
            }
            else {
                magicrule->preds[j++] =    adrule->preds[p-1];
            }
        }
        magicrule->num_preds = j;
    }
    else if (!(allargsfree( &adrule->preds[pred_index]->adorn) &&
        suppress_allfree))
    {
        magicrule->num_preds = 0;
    }
    else
        return;
    
    if (!(allargsfree( & magicrule->head->adorn) && 
            suppress_allfree))
    {
        magicrule->type = COR_MAGICRULE;
        magicrule->source_rule = adrule;
        magicrule->source_pred = adrule->preds[pred_index];

        if (magicrule->source_pred->negated ||
            corhas_grouping_args(adrule->head->args))
        {
            if (!magicrule->clause_names) {
                fputs("Allocating a new clause_names in FullMagic.C\n",stderr); 
                //    Need to create a new name
                magicrule->clause_names = new corClauseName();
                magicrule->clause_names->name = corDontCareSymbol;
                magicrule->clause_names->next = NULL; // is this needed?
            }
            magicrule->clause_names->negated = 1;
        }

        coradd_rule(magicrule, magic_rules);
        
        if (debugmagic) {
            fprintf(stderr," Magic rule:\n");
            corprint_rule(magicrule, stderr);
            fprintf(stderr,"\n");
        }
    }
    else if (debugmagic)
        fprintf(stderr,"No magic rule generated here-null rule body\n");
}

//**************************************************************************
